# Copyright 2020 Joe Milbourn <joe@milbourn.org.uk>
# Copyright 2021 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# TODO use the band field from ver_response
# TODO handle radio settings
#
# Supported features
# * Read and write memory access for 200 normal memories
# * CTCSS and DTCS for transmit and receive
# * Scan list
# * Tx off
# * Duplex (+ve, -ve, odd, and off splits)
# * Transmit power
# * Channel width (25 kHz and 12.5 kHz)
# * Retevis RT95, CRT Micron UV, and Midland DBR2500 radios
# * Full range of frequencies for tx and rx, supported band read from radio
#   during download, not verified on upload.  Radio will refuse to TX if out of
#   band.
# * Busy channel lock out (for each memory)
#
# Unsupported features
# * VFO1, VFO2, and TRF memories
# * custom CTCSS tones
# * Any non-memory radio settings
# * Reverse, talkaround, scramble
# * probably other things too - like things encoded by the unknown bits in the
#   memory struct

from chirp import chirp_common, directory, memmap, errors, util
from chirp import bitwise
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettings

import struct
import time
import logging

LOG = logging.getLogger(__name__)

# Gross hack to handle missing future module on un-updatable
# platforms like MacOS. Just avoid registering these radio
# classes for now.
try:
    from builtins import bytes
    has_future = True
except ImportError:
    has_future = False
    LOG.debug('python-future package is not '
              'available; %s requires it' % __name__)


# Here is where we define the memory map for the radio. Since
# We often just know small bits of it, we can use #seekto to skip
# around as needed.

MEM_FORMAT = '''
#seekto 0x0000;
struct {
  bbcd freq[4];
  bbcd offset[4];
  u8 unknown1;
  u8 talkaround:1,
     scramble:1,
     unknown:2,
     txpower:2,
     duplex:2;
  u8 unknown_bits1:4,
     channel_width:2,
     reverse:1,
     tx_off:1;
  u8 unknown_bits2:4,
     dtcs_decode_en:1,
     ctcss_decode_en:1,
     dtcs_encode_en:1,
     ctcss_encode_en:1;
  u8 ctcss_dec_tone;
  u8 ctcss_enc_tone;
  u8 dtcs_decode_code;
  u8 unknown_bits6:6,
     dtcs_decode_invert:1,
     dtcs_decode_code_highbit:1;
  u8 dtcs_encode_code;
  u8 unknown_bits7:6,
     dtcs_encode_invert:1,
     dtcs_encode_code_highbit:1;
  u8 unknown_bits4:6,
     busy_channel_lockout:2;
  u8 unknown6;
  u8 unknown_bits5:7,
     tone_squelch_en:1;
  u8 unknown7;
  u8 unknown8;
  u8 unknown9;
  char name[6];           // the VOX models have 6-character names
                          // the original models only have 5-character names
                          // the 1st byte is not used and will always be 0x00
  ul16 customctcss;
} memory[200];
#seekto 0x1940;
struct {
  u8 occupied_bitfield[32];
  u8 scan_enabled_bitfield[32];
} memory_status;

#seekto 0x1980;
struct {
  char line[7];           // starting display
} starting_display;

#seekto 0x1990;
struct {
  u8 code[16];            // DTMF Encode M1-M16
} pttid[16];

#seekto 0x1A90;
struct {
  u8 pttIdStart[16];      // 0x1A90 ptt id starting
  u8 pttIdEnd[16];        // 0x1AA0 ptt id ending
  u8 remoteStun[16];      // 0x1AB0 remotely stun
  u8 remoteKill[16];      // 0x1AC0 remotely kill
  u8 intervalChar;        // 0x1AD0 dtmf interval character
  u8 groupCode;           // 0x1AD1 group code
  u8 unk1ad2:6,           // 0x1AD2
     decodingResponse:2;  //        decoding response
  u8 pretime;             // 0x1AD3 pretime
  u8 firstDigitTime;      // 0x1AD4 first digit time
  u8 autoResetTime;       // 0x1AD5 auto reset time
  u8 selfID[3];           // 0x1AD6 dtmf self id
  u8 unk1ad9:7,           // 0x1AD9
     sideTone:1;          //        side tone
  u8 timeLapse;           // 0x1ADA time-lapse after encode
  u8 pauseTime;           // 0x1ADB ptt id pause time
} dtmf;

#seekto 0x3200;
struct {
  u8 unk3200:5,           // 0x3200
     beepVolume:3;        //        beep volume
  u8 unk3201:4,           // 0x3201
     frequencyStep:4;     //        frequency step
  u8 unk3202:6,           // 0x3202
     displayMode:2;       // display mode
  u8 unk0x3203;
  u8 unk3204:4,           // 0x3204
     squelchLevelA:4;     //        squelch level a
  u8 unk3205:4,           // 0x3205
     squelchLevelB:4;     //        squelch level b
  u8 unk3206:2,           // 0x3206
     speakerVol:6;        //        speaker volume
  u8 unk3207:7,           // 0x3207
     powerOnPasswd:1;     //        power-on password
  u8 unk3208:6,           // 0x3208
     scanType:2;          //        scan type
  u8 unk3209:6,           // 0x3209
     scanRecoveryT:2;     //        scan recovery time
  u8 unk320a:7,           // 0x320A
     autoPowerOn:1;       //        auto power on
  u8 unk320b:7,           // 0x320B
     main:1;              //        main
  u8 unk320c:7,           // 0x320C
     dualWatch:1;         //        dual watch (rx way select)
  u8 unk320d:5,           // 0x320D
     backlightBr:3;       //        backlight brightness
  u8 unk320e:3,           // 0x320E
     timeOutTimer:5;      //        time out timer
  u8 unk320f:6,           // 0x320F
     autoPowerOff:2;      //        auto power off
  u8 unk3210:6,           // 0x3210
     tbstFrequency:2;     //        tbst frequency
  u8 unk3211:7,           // 0x3211
     screenDir:1;         //        screen direction
  u8 unk3212:2,           // 0x3212
     micKeyBrite:6;       //        hand mic key brightness
  u8 unk3213:6,           // 0x3213
     speakerSwitch:2;     //        speaker switch
  u8 keyPA;               // 0x3214 key pa
  u8 keyPB;               // 0x3215 key pb
  u8 keyPC;               // 0x3216 key pc
  u8 keyPD;               // 0x3217 key pd
  u8 unk3218:5,           // 0x3218
     steType:3;           //        ste type
  u8 unk3219:6,           // 0x3219
     steFrequency:2;      //        ste frequency
  u8 unk321a:5,           // 0x321A
     dtmfTxTime:3;        //        dtmf transmitting time
  u8 unk_bit7_6:2,        // 0x321B
     monKeyFunction:1,    //        mon key function
     channelLocked:1,     //        channel locked
     saveChParameter:1,   //        save channel parameter
     powerOnReset:1,      //        power on reset
     trfEnable:1,         //        trf enable
     knobMode:1;          //        knob mode
  u8 unk321c:7,           // 0x321C
     voxOnOff:1;          //        vox on/off
  u8 voxLevel;            // 0x321D vox level
  u8 voxDelay;            // 0x321E vox delay
} settings;

#seekto 0x3240;
struct {
  char digits[6];         // password
} password;

#seekto 0x3250;
struct {
  u8 keyMode1P1;          // 0x3250 key mode 1 p1
  u8 keyMode1P2;          // 0x3251 key mode 1 p2
  u8 keyMode1P3;          // 0x3252 key mode 1 p3
  u8 keyMode1P4;          // 0x3253 key mode 1 p4
  u8 keyMode1P5;          // 0x3254 key mode 1 p5
  u8 keyMode1P6;          // 0x3255 key mode 1 p6
  u8 keyMode2P1;          // 0x3256 key mode 2 p1
  u8 keyMode2P2;          // 0x3257 key mode 2 p2
  u8 keyMode2P3;          // 0x3258 key mode 2 p3
  u8 keyMode2P4;          // 0x3259 key mode 2 p4
  u8 keyMode2P5;          // 0x325A key mode 2 p5
  u8 keyMode2P6;          // 0x325B key mode 2 p6
} pfkeys;

#seekto 0x3260;
struct {
  u8 mrChanA;             // 0x3260 mr channel a
  u8 unknown1_0:7,        // 0x3261
     vfomrA:1;            //        vfo/mr mode a
  u8 unknown2;
  u8 unknown3;
  u8 unknown4;
  u8 unknown5;
  u8 unknown6;
  u8 mrChanB;             // 0x3267 mr channel b
  u8 unknown8_0:4,        // 0x3268
     scan_active:1,
     unknown8_1:2,
     vfomrB:1;            //        vfo/mr mode b
  u8 unknown9;
  u8 unknowna;
  u8 unknownb;
  u8 unknownc;
  u8 bandlimit;           // 0x326D mode
  u8 unknownd;
  u8 unknowne;
  u8 unknownf;
} radio_settings;
'''

# Format for the version messages returned by the radio
VER_FORMAT = '''
u8 hdr;
char model[7];
u8 bandlimit;
char version[6];
u8 ack;
'''

TXPOWER_LOW = 0x00
TXPOWER_MED = 0x01
TXPOWER_HIGH = 0x02

DUPLEX_NOSPLIT = 0x00
DUPLEX_POSSPLIT = 0x01
DUPLEX_NEGSPLIT = 0x02
DUPLEX_ODDSPLIT = 0x03

CHANNEL_WIDTH_25kHz = 0x02
CHANNEL_WIDTH_20kHz = 0x01
CHANNEL_WIDTH_12d5kHz = 0x00

BUSY_CHANNEL_LOCKOUT_OFF = 0x00
BUSY_CHANNEL_LOCKOUT_REPEATER = 0x01
BUSY_CHANNEL_LOCKOUT_BUSY = 0x02

MEMORY_ADDRESS_RANGE = (0x0000, 0x3290)
MEMORY_RW_BLOCK_SIZE = 0x10
MEMORY_RW_BLOCK_CMD_SIZE = 0x16

POWER_LEVELS = [chirp_common.PowerLevel('Low', dBm=37),
                chirp_common.PowerLevel('Medium', dBm=40),
                chirp_common.PowerLevel('High', dBm=44)]

# CTCSS Tone definitions
TONE_CUSTOM_CTCSS = 0x33
TONE_MAP_VAL_TO_TONE = {0x00: 62.5, 0x01: 67.0, 0x02: 69.3,
                        0x03: 71.9, 0x04: 74.4, 0x05: 77.0,
                        0x06: 79.7, 0x07: 82.5, 0x08: 85.4,
                        0x09: 88.5, 0x0a: 91.5, 0x0b: 94.8,
                        0x0c: 97.4, 0x0d: 100.0, 0x0e: 103.5,
                        0x0f: 107.2, 0x10: 110.9, 0x11: 114.8,
                        0x12: 118.8, 0x13: 123.0, 0x14: 127.3,
                        0x15: 131.8, 0x16: 136.5, 0x17: 141.3,
                        0x18: 146.2, 0x19: 151.4, 0x1a: 156.7,
                        0x1b: 159.8, 0x1c: 162.2, 0x1d: 165.5,
                        0x1e: 167.9, 0x1f: 171.3, 0x20: 173.8,
                        0x21: 177.3, 0x22: 179.9, 0x23: 183.5,
                        0x24: 186.2, 0x25: 189.9, 0x26: 192.8,
                        0x27: 196.6, 0x28: 199.5, 0x29: 203.5,
                        0x2a: 206.5, 0x2b: 210.7, 0x2c: 218.1,
                        0x2d: 225.7, 0x2e: 229.1, 0x2f: 233.6,
                        0x30: 241.8, 0x31: 250.3, 0x32: 254.1}

TONE_MAP_TONE_TO_VAL = {TONE_MAP_VAL_TO_TONE[val]: val
                        for val in TONE_MAP_VAL_TO_TONE}

TONES_EN_TXTONE = (1 << 3)
TONES_EN_RXTONE = (1 << 2)
TONES_EN_TXCODE = (1 << 1)
TONES_EN_RXCODE = (1 << 0)
TONES_EN_NO_TONE = 0

# Radio supports upper case and symbols
CHARSET_ASCII_PLUS = chirp_common.CHARSET_UPPER_NUMERIC + '- '

# Band limits as defined by the band byte in ver_response, defined in Hz, for
# VHF and UHF, used for RX and TX.
BAND_LIMITS = {0x00: [(144000000, 148000000), (430000000, 440000000)],
               0x01: [(136000000, 174000000), (400000000, 490000000)],
               0x02: [(144000000, 146000000), (430000000, 440000000)]}


# Get band limits from a band limit value
def get_band_limits_Hz(limit_value):
    if limit_value not in BAND_LIMITS:
        limit_value = 0x01
        LOG.warning('Unknown band limit value 0x%02x, default to 0x01')
    bandlimitfrequencies = BAND_LIMITS[limit_value]
    return bandlimitfrequencies


# Calculate the checksum used in serial packets
def checksum(message_bytes):
    mask = 0xFF
    checksum = 0
    for b in message_bytes:
        checksum = (checksum + b) & mask
    return checksum


# Send a command to the radio, return any reply stripping the echo of the
# command (tx and rx share a single pin in this radio)
def send_serial_command(serial, command, expectedlen=None):
    ''' send a command to the radio, and return any response.
    set expectedlen to return as soon as that many bytes are read.
    '''
    serial.write(command)
    serial.flush()

    response = b''
    tout = time.time() + 0.5
    while time.time() < tout:
        if serial.inWaiting():
            response += serial.read()
        # remember everything gets echo'd back
        if len(response) - len(command) == expectedlen:
            break

    # cut off what got echo'd back, we don't need to see it again
    if response.startswith(command):
        response = response[len(command):]

    return response


# strip trailing 0x00 to convert a string returned by bitwise.parse into a
# python string
def cstring_to_py_string(cstring):
    return "".join(c for c in cstring if c != '\x00')


# Check the radio version reported to see if it's one we support,
# returns bool version supported, and the band index
def check_ver(ver_response, allowed_types):
    ''' Check the returned radio version is one we approve of '''

    LOG.debug('ver_response = ')
    LOG.debug(util.hexprint(ver_response))

    global HAS_VOX

    resp = bitwise.parse(VER_FORMAT, ver_response)
    verok = False

    if resp.hdr == 0x49 and resp.ack == 0x06:
        model, version = [cstring_to_py_string(bitwise.get_string(s)).strip()
                          for s in (resp.model, resp.version)]
        LOG.debug('radio model: \'%s\' version: \'%s\'' %
                  (model, version))
        LOG.debug('allowed_types = %s' % allowed_types)

        HAS_VOX = "P" == model[-1:]

        if model in allowed_types:
            LOG.debug('model in allowed_types')

            if version in allowed_types[model]:
                LOG.debug('version in allowed_types[model]')
                verok = True
    else:
        raise errors.RadioError('Failed to parse version response')

    return verok, int(resp.bandlimit)


# Put the radio in programming mode, sending the initial command and checking
# the response.  raise RadioError if there is no response (500ms timeout), and
# if the returned version isn't matched by check_ver
def enter_program_mode(radio):
    serial = radio.pipe
    # place the radio in program mode, and confirm
    program_response = send_serial_command(serial, b'PROGRAM')

    if program_response != b'QX\x06':
        raise errors.RadioError('No initial response from radio.')
    LOG.debug('entered program mode')

    # read the radio ID string, make sure it matches one we know about
    ver_response = send_serial_command(serial, b'\x02')

    verok, bandlimit = check_ver(ver_response, radio.ALLOWED_RADIO_TYPES)
    if not verok:
        exit_program_mode(radio)
        ver = "V2" if radio.VENDOR == "CRT" else "VOX"
        if HAS_VOX and ("V2" and "VOX") not in radio.MODEL:
            raise errors.RadioError(
                'Radio identified as model with VOX\n'
                'Try %s-%s %s' %
                (radio.VENDOR, radio.MODEL, ver))
        else:
            raise errors.RadioError(
                'Radio version not in allowed list for %s-%s:\n'
                '%s' %
                (radio.VENDOR, radio.MODEL, util.hexprint(ver_response)))

    return bandlimit


# Exit programming mode
def exit_program_mode(radio):
    send_serial_command(radio.pipe, b'END')


# Parse a packet from the radio returning the header (R/W, address, data, and
# checksum valid
def parse_read_response(resp):
    addr = resp[:4]
    data = bytes(resp[4:-2])
    cs = checksum(d for d in resp[1:-2])
    valid = cs == resp[-2]
    if not valid:
        LOG.error('checksumfail: %02x, expected %02x' % (cs, resp[-2]))
        LOG.error('msg data: %s' % util.hexprint(resp))
    return addr, data, valid


# Download data from the radio and populate the memory map
def do_download(radio):
    '''Download memories from the radio'''

    # Get the serial port connection
    serial = radio.pipe

    try:
        enter_program_mode(radio)

        memory_data = bytes()

        # status info for the UI
        status = chirp_common.Status()
        status.cur = 0
        status.max = (MEMORY_ADDRESS_RANGE[1] -
                      MEMORY_ADDRESS_RANGE[0]) // MEMORY_RW_BLOCK_SIZE
        status.msg = 'Cloning from radio...'
        radio.status_fn(status)

        for addr in range(MEMORY_ADDRESS_RANGE[0],
                          MEMORY_ADDRESS_RANGE[1] + MEMORY_RW_BLOCK_SIZE,
                          MEMORY_RW_BLOCK_SIZE):
            read_command = struct.pack('>BHB', 0x52, addr,
                                       MEMORY_RW_BLOCK_SIZE)
            read_response = send_serial_command(serial, read_command,
                                                MEMORY_RW_BLOCK_CMD_SIZE)
            # LOG.debug('read response:\n%s' % util.hexprint(read_response))

            address, data, valid = parse_read_response(read_response)
            memory_data += data

            # update UI
            status.cur = (addr - MEMORY_ADDRESS_RANGE[0])\
                // MEMORY_RW_BLOCK_SIZE
            radio.status_fn(status)

        exit_program_mode(radio)
    except errors.RadioError as e:
        raise e
    except Exception as e:
        raise errors.RadioError('Failed to download from radio: %s' % e)

    return memmap.MemoryMapBytes(memory_data)


# Build a write data command to send to the radio
def make_write_data_cmd(addr, data, datalen):
    cmd = struct.pack('>BHB', 0x57, addr, datalen)
    cmd += data
    cs = checksum(c for c in cmd[1:])
    cmd += struct.pack('>BB', cs, 0x06)
    return cmd


# Upload a memory map to the radio
def do_upload(radio):
    try:
        bandlimit = enter_program_mode(radio)

        if bandlimit != radio._memobj.radio_settings.bandlimit:
            LOG.warning('radio and image bandlimits differ'
                        ' some channels many not work'
                        ' (img:0x%02x radio:0x%02x)' %
                        (int(bandlimit),
                         int(radio._memobj.radio_settings.bandlimit)))
            LOG.warning('radio bands: %s' % get_band_limits_Hz(
                         int(radio._memobj.radio_settings.bandlimit)))
            LOG.warning('img bands: %s' % get_band_limits_Hz(bandlimit))

        serial = radio.pipe

        # send the initial message, radio responds with something that looks a
        # bit like a bitfield, but I don't know what it is yet.
        read_command = struct.pack('>BHB', 0x52, 0x3b10, MEMORY_RW_BLOCK_SIZE)
        read_response = send_serial_command(serial, read_command,
                                            MEMORY_RW_BLOCK_CMD_SIZE)
        address, data, valid = parse_read_response(read_response)
        LOG.debug('Got initial response from radio: %s' %
                  util.hexprint(read_response))

        bptr = 0

        memory_addrs = list(range(MEMORY_ADDRESS_RANGE[0],
                                  MEMORY_ADDRESS_RANGE[1] +
                                  MEMORY_RW_BLOCK_SIZE,
                                  MEMORY_RW_BLOCK_SIZE))

        # status info for the UI
        status = chirp_common.Status()
        status.cur = 0
        status.max = len(memory_addrs)
        status.msg = 'Cloning to radio...'
        radio.status_fn(status)

        for idx, addr in enumerate(memory_addrs):
            write_command = make_write_data_cmd(
                addr, radio._mmap[bptr:bptr+MEMORY_RW_BLOCK_SIZE],
                MEMORY_RW_BLOCK_SIZE)
            # LOG.debug('write data:\n%s' % util.hexprint(write_command))
            write_response = send_serial_command(serial, write_command, 0x01)
            bptr += MEMORY_RW_BLOCK_SIZE

            if write_response == '\x0a':
                # NACK from radio, e.g. checksum wrong
                LOG.debug('Radio returned 0x0a - NACK:')
                LOG.debug(' * write cmd:\n%s' % util.hexprint(write_command))
                LOG.debug(' * write response:\n%s' %
                          util.hexprint(write_response))
                exit_program_mode(radio)
                raise errors.RadioError('Radio NACK\'d write command')

            # update UI
            status.cur = idx
            radio.status_fn(status)
        exit_program_mode(radio)
    except errors.RadioError:
        raise
    except Exception as e:
        raise errors.RadioError('Failed to download from radio: %s' % e)


# Get the value of @bitfield @number of bits in from 0
def get_bitfield(bitfield, number):
    ''' Get the value of @bitfield @number of bits in '''
    byteidx = number//8
    bitidx = number - (byteidx * 8)
    return bitfield[byteidx] & (1 << bitidx)


# Set the @value of @bitfield @number of bits in from 0
def set_bitfield(bitfield, number, value):
    ''' Set the @value of @bitfield @number of bits in '''
    byteidx = number//8
    bitidx = number - (byteidx * 8)
    if value is True:
        bitfield[byteidx] |= (1 << bitidx)
    else:
        bitfield[byteidx] &= ~(1 << bitidx)
    return bitfield


# Translate the radio's version of a code as stored to a real code
def dtcs_code_bits_to_val(highbit, lowbyte):
    return chirp_common.ALL_DTCS_CODES[highbit*256 + lowbyte]


# Translate the radio's version of a tone as stored to a real tone
def ctcss_tone_bits_to_val(tone_byte):
    # TODO use the custom setting 0x33 and ref the custom ctcss
    # field
    tone_byte = int(tone_byte)
    if tone_byte in TONE_MAP_VAL_TO_TONE:
        return TONE_MAP_VAL_TO_TONE[tone_byte]
    elif tone_byte == TONE_CUSTOM_CTCSS:
        LOG.info('custom ctcss not implemented (yet?).')
    else:
        raise errors.UnsupportedToneError('unknown ctcss tone value: %02x' %
                                          tone_byte)


# Translate a real tone to the radio's version as stored
def ctcss_code_val_to_bits(tone_value):
    if tone_value in TONE_MAP_TONE_TO_VAL:
        return TONE_MAP_TONE_TO_VAL[tone_value]
    else:
        raise errors.UnsupportedToneError('Tone %f not supported' % tone_value)


# Translate a real code to the radio's version as stored
def dtcs_code_val_to_bits(code):
    val = chirp_common.ALL_DTCS_CODES.index(code)
    return (val & 0xFF), ((val >> 8) & 0x01)


class AnyTone778UVBase(chirp_common.CloneModeRadio,
                       chirp_common.ExperimentalRadio):
    '''AnyTone 778UV and probably Retevis RT95 and others'''
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    NAME_LENGTH = 5
    HAS_VOX = False

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()

        rp.experimental = \
            ('This is experimental support for the %s %s.  '
             'Please send in bug and enhancement requests!' %
             (cls.VENDOR, cls.MODEL))

        return rp

    # Return information about this radio's features, including
    # how many memories it has, what bands it supports, etc
    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False
        rf.has_settings = True
        rf.can_odd_split = True
        rf.has_name = True
        rf.has_offset = True
        rf.valid_name_length = self.NAME_LENGTH
        rf.valid_duplexes = ['', '+', '-', 'split', 'off']
        rf.valid_characters = CHARSET_ASCII_PLUS

        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.valid_dtcs_codes = chirp_common.ALL_DTCS_CODES
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = ['Tone->Tone',
                                'Tone->DTCS',
                                'DTCS->Tone',
                                'DTCS->DTCS',
                                'DTCS->',
                                '->DTCS',
                                '->Tone']

        rf.memory_bounds = (1, 200)  # This radio supports memories 1-200
        try:
            rf.valid_bands = get_band_limits_Hz(
                int(self._memobj.radio_settings.bandlimit))
        except AttributeError:
            # If we're asked without memory loaded, assume the most permissive
            rf.valid_bands = get_band_limits_Hz(1)
        except Exception as e:
            LOG.exception(
                'Failed to get band limits for anytone778uv: %s' % e)
            rf.valid_bands = get_band_limits_Hz(1)
        rf.valid_modes = ['FM', 'NFM']
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tuning_steps = [2.5, 5, 6.25, 10, 12.5, 20, 25, 30, 50]
        rf.has_tuning_step = False
        return rf

    # Do a download of the radio from the serial port
    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    # Do an upload of the radio to the serial port
    def sync_out(self):
        do_upload(self)

    # Convert the raw byte array into a memory object structure
    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    # Return a raw representation of the memory object, which
    # is very helpful for development
    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    # Extract a high-level memory object from the low-level memory map
    # This is called to populate a memory in the UI
    def get_memory(self, number):
        number -= 1
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[number]
        _mem_status = self._memobj.memory_status

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()
        mem.number = number + 1           # Set the memory number

        # Check if this memory is present in the occupied list
        mem.empty = get_bitfield(_mem_status.occupied_bitfield, number) == 0

        if not mem.empty:
            # Check if this memory is in the scan enabled list
            mem.skip = ''
            if get_bitfield(_mem_status.scan_enabled_bitfield, number) == 0:
                mem.skip = 'S'

            # set the name
            if self.NAME_LENGTH == 5:
                # original models with 5-character name length
                temp_name = str(_mem.name)
                # Strip the first character and set the alpha tag
                mem.name = str(temp_name[1:6]).rstrip()
            else:
                # new VOX models with 6-character name length
                mem.name = str(_mem.name).rstrip()  # Set the alpha tag

            # Convert your low-level frequency and offset to Hertz
            mem.freq = int(_mem.freq) * 10
            mem.offset = int(_mem.offset) * 10

            # Set the duplex flags
            if _mem.duplex == DUPLEX_POSSPLIT:
                mem.duplex = '+'
            elif _mem.duplex == DUPLEX_NEGSPLIT:
                mem.duplex = '-'
            elif _mem.duplex == DUPLEX_NOSPLIT:
                mem.duplex = ''
            elif _mem.duplex == DUPLEX_ODDSPLIT:
                mem.duplex = 'split'
            else:
                LOG.error('%s: get_mem: unhandled duplex: %02x' %
                          (mem.name, _mem.duplex))

            # handle tx off
            if _mem.tx_off:
                mem.duplex = 'off'

            # Set the channel width
            if _mem.channel_width == CHANNEL_WIDTH_25kHz:
                mem.mode = 'FM'
            elif _mem.channel_width == CHANNEL_WIDTH_20kHz:
                LOG.info(
                    '%s: get_mem: promoting 20 kHz channel width to 25 kHz' %
                    mem.name)
                mem.mode = 'FM'
            elif _mem.channel_width == CHANNEL_WIDTH_12d5kHz:
                mem.mode = 'NFM'
            else:
                LOG.error('%s: get_mem: unhandled channel width: 0x%02x' %
                          (mem.name, _mem.channel_width))

            # set the power level
            if _mem.txpower == TXPOWER_LOW:
                mem.power = POWER_LEVELS[0]
            elif _mem.txpower == TXPOWER_MED:
                mem.power = POWER_LEVELS[1]
            elif _mem.txpower == TXPOWER_HIGH:
                mem.power = POWER_LEVELS[2]
            else:
                LOG.error('%s: get_mem: unhandled power level: 0x%02x' %
                          (mem.name, _mem.txpower))

            # CTCSS Tones
            # TODO support custom ctcss tones here
            txtone = None
            rxtone = None
            rxcode = None
            txcode = None

            # check if dtcs tx is enabled
            if _mem.dtcs_encode_en:
                txcode = dtcs_code_bits_to_val(_mem.dtcs_encode_code_highbit,
                                               _mem.dtcs_encode_code)

            # check if dtcs rx is enabled
            if _mem.dtcs_decode_en:
                rxcode = dtcs_code_bits_to_val(_mem.dtcs_decode_code_highbit,
                                               _mem.dtcs_decode_code)

            if txcode is not None:
                LOG.debug('%s: get_mem dtcs_enc: %d' % (mem.name, txcode))
            if rxcode is not None:
                LOG.debug('%s: get_mem dtcs_dec: %d' % (mem.name, rxcode))

            # tsql set if radio squelches on tone
            tsql = _mem.tone_squelch_en

            # check if ctcss tx is enabled
            if _mem.ctcss_encode_en:
                txtone = ctcss_tone_bits_to_val(_mem.ctcss_enc_tone)

            # check if ctcss rx is enabled
            if _mem.ctcss_decode_en:
                rxtone = ctcss_tone_bits_to_val(_mem.ctcss_dec_tone)

            # Define this here to allow a readable if-else tree enabling tone
            # options
            enabled = 0
            enabled |= (txtone is not None) * TONES_EN_TXTONE
            enabled |= (rxtone is not None) * TONES_EN_RXTONE
            enabled |= (txcode is not None) * TONES_EN_TXCODE
            enabled |= (rxcode is not None) * TONES_EN_RXCODE

            # Add some debugging output for the tone bitmap
            enstr = []
            if enabled & TONES_EN_TXTONE:
                enstr += ['TONES_EN_TXTONE']
            if enabled & TONES_EN_RXTONE:
                enstr += ['TONES_EN_RXTONE']
            if enabled & TONES_EN_TXCODE:
                enstr += ['TONES_EN_TXCODE']
            if enabled & TONES_EN_RXCODE:
                enstr += ['TONES_EN_RXCODE']
            if enabled == 0:
                enstr = ['TONES_EN_NOTONE']
            LOG.debug('%s: enabled = %s' % (
                mem.name, '|'.join(enstr)))

            mem.tmode = ''
            if enabled == TONES_EN_NO_TONE:
                mem.tmode = ''
            elif enabled == TONES_EN_TXTONE:
                mem.tmode = 'Tone'
                mem.rtone = txtone
            elif enabled == TONES_EN_RXTONE and tsql:
                mem.tmode = 'Cross'
                mem.cross_mode = '->Tone'
                mem.ctone = rxtone
            elif enabled == (TONES_EN_TXTONE | TONES_EN_RXTONE) and tsql:
                if txtone == rxtone:  # TSQL
                    mem.tmode = 'TSQL'
                    mem.ctone = txtone
                else:  # Tone->Tone
                    mem.tmode = 'Cross'
                    mem.cross_mode = 'Tone->Tone'
                    mem.ctone = rxtone
                    mem.rtone = txtone
            elif enabled == TONES_EN_TXCODE:
                mem.tmode = 'Cross'
                mem.cross_mode = 'DTCS->'
                mem.dtcs = txcode
            elif enabled == TONES_EN_RXCODE and tsql:
                mem.tmode = 'Cross'
                mem.cross_mode = '->DTCS'
                mem.rx_dtcs = rxcode
            elif enabled == (TONES_EN_TXCODE | TONES_EN_RXCODE) and tsql:
                if rxcode == txcode:
                    mem.tmode = 'DTCS'
                    mem.rx_dtcs = rxcode
                    # #8327 Not sure this is the correct interpretation of
                    # DevelopersToneModes, but it seems to make it work round
                    # tripping with the Anytone software.  DTM implies that we
                    # might not need to set mem.dtcs, but if we do it only DTCS
                    # rx works (as if we were Cross:None->DTCS).
                    mem.dtcs = rxcode
                else:
                    mem.tmode = 'Cross'
                    mem.cross_mode = 'DTCS->DTCS'
                    mem.rx_dtcs = rxcode
                    mem.dtcs = txcode
            elif enabled == (TONES_EN_TXCODE | TONES_EN_RXTONE) and tsql:
                mem.tmode = 'Cross'
                mem.cross_mode = 'DTCS->Tone'
                mem.dtcs = txcode
                mem.ctone = rxtone
            elif enabled == (TONES_EN_TXTONE | TONES_EN_RXCODE) and tsql:
                mem.tmode = 'Cross'
                mem.cross_mode = 'Tone->DTCS'
                mem.rx_dtcs = rxcode
                mem.rtone = txtone
            else:
                LOG.error('%s: Unhandled tmode enabled = %d.' % (
                    mem.name, enabled))

                # Can get here if e.g. TONE_EN_RXCODE is set and tsql isn't
                # In that case should we perhaps store the tone and code values
                # if they're present and then setup tmode and cross_mode as
                # appropriate later?

            # set the dtcs polarity
            dtcs_pol_bit_to_str = {0: 'N', 1: 'R'}
            mem.dtcs_polarity = '%s%s' %\
                (dtcs_pol_bit_to_str[_mem.dtcs_encode_invert == 1],
                 dtcs_pol_bit_to_str[_mem.dtcs_decode_invert == 1])

            # Extra
            mem.extra = RadioSettingGroup("Extra", "extra")

            # Busy channel lockout
            bcl_options = ['Off', 'Repeater', 'Busy']
            bcl_option = bcl_options[_mem.busy_channel_lockout]
            rs = RadioSetting("busy_channel_lockout", "Busy Channel Lockout",
                              RadioSettingValueList(bcl_options, bcl_option))
            mem.extra.append(rs)

        return mem

    # Store details about a high-level memory to the memory map
    # This is called when a user edits a memory in the UI
    def set_memory(self, mem):
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]
        _mem_status = self._memobj.memory_status

        # set the occupied bitfield
        _mem_status.occupied_bitfield = \
            set_bitfield(_mem_status.occupied_bitfield, mem.number - 1,
                         not mem.empty)

        # set the scan add bitfield
        _mem_status.scan_enabled_bitfield = \
            set_bitfield(_mem_status.scan_enabled_bitfield, mem.number - 1,
                         (not mem.empty) and (mem.skip != 'S'))

        if mem.empty:
            # Set the whole memory to 0xff
            _mem.set_raw('\xff' * (_mem.size() // 8))
        else:
            _mem.set_raw('\x00' * (_mem.size() // 8))

            _mem.freq = int(mem.freq / 10)
            _mem.offset = int(mem.offset / 10)

            # Store the alpha tag
            if self.NAME_LENGTH == 5:
                # original models with 5-character name length
                temp_name = mem.name.ljust(self.NAME_LENGTH)[:self.NAME_LENGTH]
                # prefix the 5-character name with 0x00 to fit structure
                _mem.name = temp_name.rjust(6, chr(00))
            else:
                # new VOX models with 6-character name length
                _mem.name = mem.name.ljust(self.NAME_LENGTH)[:self.NAME_LENGTH]

            # Set duplex bitfields
            if mem.duplex == '+':
                _mem.duplex = DUPLEX_POSSPLIT
            elif mem.duplex == '-':
                _mem.duplex = DUPLEX_NEGSPLIT
            elif mem.duplex == '':
                _mem.duplex = DUPLEX_NOSPLIT
            elif mem.duplex == 'split':
                # TODO: this is an unverified punt!
                _mem.duplex = DUPLEX_ODDSPLIT
            else:
                LOG.error('%s: set_mem: unhandled duplex: %s' %
                          (mem.name, mem.duplex))

            # handle tx off
            _mem.tx_off = 0
            if mem.duplex == 'off':
                _mem.tx_off = 1

            # Set the channel width - remember we promote 20 kHz channels to FM
            # on import
            # , so don't handle them here
            if mem.mode == 'FM':
                _mem.channel_width = CHANNEL_WIDTH_25kHz
            elif mem.mode == 'NFM':
                _mem.channel_width = CHANNEL_WIDTH_12d5kHz
            else:
                LOG.error('%s: set_mem: unhandled mode: %s' % (
                    mem.name, mem.mode))

            # set the power level
            if mem.power == POWER_LEVELS[0]:
                _mem.txpower = TXPOWER_LOW
            elif mem.power == POWER_LEVELS[1]:
                _mem.txpower = TXPOWER_MED
            elif mem.power == POWER_LEVELS[2]:
                _mem.txpower = TXPOWER_HIGH
            else:
                LOG.error('%s: set_mem: unhandled power level: %s' %
                          (mem.name, mem.power))

            # TODO set the CTCSS values
            # TODO support custom ctcss tones here
            # Default - tones off, carrier sql
            _mem.ctcss_encode_en = 0
            _mem.ctcss_decode_en = 0
            _mem.tone_squelch_en = 0
            _mem.ctcss_enc_tone = 0x00
            _mem.ctcss_dec_tone = 0x00
            _mem.customctcss = 0x00
            _mem.dtcs_encode_en = 0
            _mem.dtcs_encode_code_highbit = 0
            _mem.dtcs_encode_code = 0
            _mem.dtcs_encode_invert = 0
            _mem.dtcs_decode_en = 0
            _mem.dtcs_decode_code_highbit = 0
            _mem.dtcs_decode_code = 0
            _mem.dtcs_decode_invert = 0

            dtcs_pol_str_to_bit = {'N': 0, 'R': 1}
            _mem.dtcs_encode_invert = dtcs_pol_str_to_bit[mem.dtcs_polarity[0]]
            _mem.dtcs_decode_invert = dtcs_pol_str_to_bit[mem.dtcs_polarity[1]]

            if mem.tmode == 'Tone':
                _mem.ctcss_encode_en = 1
                _mem.ctcss_enc_tone = ctcss_code_val_to_bits(mem.rtone)
            elif mem.tmode == 'TSQL':
                _mem.ctcss_encode_en = 1
                _mem.ctcss_enc_tone = ctcss_code_val_to_bits(mem.ctone)
                _mem.ctcss_decode_en = 1
                _mem.tone_squelch_en = 1
                _mem.ctcss_dec_tone = ctcss_code_val_to_bits(mem.ctone)
            elif mem.tmode == 'DTCS':
                _mem.dtcs_encode_en = 1
                _mem.dtcs_encode_code, _mem.dtcs_encode_code_highbit = \
                    dtcs_code_val_to_bits(mem.rx_dtcs)
                _mem.dtcs_decode_en = 1
                _mem.dtcs_decode_code, _mem.dtcs_decode_code_highbit = \
                    dtcs_code_val_to_bits(mem.rx_dtcs)
                _mem.tone_squelch_en = 1
            elif mem.tmode == 'Cross':
                txmode, rxmode = mem.cross_mode.split('->')

                if txmode == 'Tone':
                    _mem.ctcss_encode_en = 1
                    _mem.ctcss_enc_tone = ctcss_code_val_to_bits(mem.rtone)
                elif txmode == '':
                    pass
                elif txmode == 'DTCS':
                    _mem.dtcs_encode_en = 1
                    _mem.dtcs_encode_code, _mem.dtcs_encode_code_highbit = \
                        dtcs_code_val_to_bits(mem.dtcs)
                else:
                    LOG.error('%s: unhandled cross TX mode: %s' % (
                        mem.name, mem.cross_mode))

                if rxmode == 'Tone':
                    _mem.ctcss_decode_en = 1
                    _mem.tone_squelch_en = 1
                    _mem.ctcss_dec_tone = ctcss_code_val_to_bits(mem.ctone)
                elif rxmode == '':
                    pass
                elif rxmode == 'DTCS':
                    _mem.dtcs_decode_en = 1
                    _mem.dtcs_decode_code, _mem.dtcs_decode_code_highbit = \
                        dtcs_code_val_to_bits(mem.rx_dtcs)
                    _mem.tone_squelch_en = 1
                else:
                    LOG.error('%s: unhandled cross RX mode: %s' % (
                        mem.name, mem.cross_mode))
            else:
                LOG.error('%s: Unhandled tmode/cross %s/%s.' %
                          (mem.name, mem.tmode, mem.cross_mode))
            LOG.debug('%s: tmode=%s, cross=%s, rtone=%f, ctone=%f' % (
                mem.name, mem.tmode, mem.cross_mode, mem.rtone, mem.ctone))
            LOG.debug('%s: CENC=%d, CDEC=%d, t(enc)=%02x, t(dec)=%02x' % (
                mem.name,
                _mem.ctcss_encode_en,
                _mem.ctcss_decode_en,
                ctcss_code_val_to_bits(mem.rtone),
                ctcss_code_val_to_bits(mem.ctone)))

            # set unknown defaults, based on reading memory set by vendor tool
            _mem.unknown1 = 0x00
            _mem.unknown6 = 0x00
            _mem.unknown7 = 0x00
            _mem.unknown8 = 0x00
            _mem.unknown9 = 0x00

            # Extra settings (eg. Busy channel lockout)
            for setting in mem.extra:
                setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        """Translate the MEM_FORMAT structs into setstuf in the UI"""
        _settings = self._memobj.settings
        _radio_settings = self._memobj.radio_settings
        _password = self._memobj.password
        _pfkeys = self._memobj.pfkeys
        _dtmf = self._memobj.dtmf

        # Function Setup
        function = RadioSettingGroup("function", "Function Setup")
        group = RadioSettings(function)

        # MODE SET
        # Channel Locked
        rs = RadioSettingValueBoolean(_settings.channelLocked)
        rset = RadioSetting("settings.channelLocked", "Channel locked", rs)
        function.append(rset)

        # Menu 3 - Display Mode
        options = ["Frequency", "Channel", "Name"]
        rs = RadioSettingValueList(options, options[_settings.displayMode])
        rset = RadioSetting("settings.displayMode", "Display Mode", rs)
        function.append(rset)

        # VFO/MR A
        options = ["MR", "VFO"]
        rs = RadioSettingValueList(options, options[_radio_settings.vfomrA])
        rset = RadioSetting("radio_settings.vfomrA", "VFO/MR mode A", rs)
        function.append(rset)

        # MR Channel A
        options = ["%s" % x for x in range(1, 201)]
        rs = RadioSettingValueList(options, options[_radio_settings.mrChanA])
        rset = RadioSetting("radio_settings.mrChanA", "MR channel A", rs)
        function.append(rset)

        # VFO/MR B
        options = ["MR", "VFO"]
        rs = RadioSettingValueList(options, options[_radio_settings.vfomrB])
        rset = RadioSetting("radio_settings.vfomrB", "VFO/MR mode B", rs)
        function.append(rset)

        # MR Channel B
        options = ["%s" % x for x in range(1, 201)]
        rs = RadioSettingValueList(options, options[_radio_settings.mrChanB])
        rset = RadioSetting("radio_settings.mrChanB", "MR channel B", rs)
        function.append(rset)

        # DISPLAY SET
        # Starting Display
        name = ""
        for i in range(7):  # 0 - 7
            name += chr(int(self._memobj.starting_display.line[i]))
        name = name.upper().rstrip()  # remove trailing spaces

        rs = RadioSettingValueString(0, 7, name)
        rs.set_charset(chirp_common.CHARSET_ALPHANUMERIC)
        rset = RadioSetting("starting_display.line", "Starting display", rs)
        function.append(rset)

        # Menu 11 - Backlight Brightness
        options = ["%s" % x for x in range(1, 4)]
        rs = RadioSettingValueList(options, options[_settings.backlightBr - 1])
        rset = RadioSetting("settings.backlightBr", "Backlight brightness", rs)
        function.append(rset)

        # Menu 15 - Screen Direction
        options = ["Positive", "Inverted"]
        rs = RadioSettingValueList(options, options[_settings.screenDir])
        rset = RadioSetting("settings.screenDir", "Screen direction", rs)
        function.append(rset)

        # Hand Mic Key Brightness
        options = ["%s" % x for x in range(1, 32)]
        rs = RadioSettingValueList(options, options[_settings.micKeyBrite - 1])
        rset = RadioSetting("settings.micKeyBrite",
                            "Hand mic key brightness", rs)
        function.append(rset)

        # VOL SET
        # Menu 1 - Beep Volume
        options = ["OFF"] + ["%s" % x for x in range(1, 6)]
        rs = RadioSettingValueList(options, options[_settings.beepVolume])
        rset = RadioSetting("settings.beepVolume", "Beep volume", rs)
        function.append(rset)

        # Menu 5 - Volume level Setup
        options = ["%s" % x for x in range(1, 37)]
        rs = RadioSettingValueList(options, options[_settings.speakerVol - 1])
        rset = RadioSetting("settings.speakerVol", "Speaker volume", rs)
        function.append(rset)

        # Menu 16 - Speaker Switch
        options = ["Host on | Hand mic off", "Host on | Hand mic on",
                   "Host off | Hand mic on"]
        rs = RadioSettingValueList(options, options[_settings.speakerSwitch])
        rset = RadioSetting("settings.speakerSwitch", "Speaker switch", rs)
        function.append(rset)

        # STE SET
        # STE Frequency
        options = ["Off", "55.2 Hz", "259.2 Hz"]
        rs = RadioSettingValueList(options, options[_settings.steFrequency])
        rset = RadioSetting("settings.steFrequency", "STE frequency", rs)
        function.append(rset)

        # STE Type
        options = ["Off", "Silent", "120 degrees", "180 degrees",
                   "240 degrees"]
        rs = RadioSettingValueList(options, options[_settings.steType])
        rset = RadioSetting("settings.steType", "STE type", rs)
        function.append(rset)

        # ON/OFF SET
        # The Power-on Password feature is not available on models with VOX
        if not self.HAS_VOX:
            # Power-on Password
            rs = RadioSettingValueBoolean(_settings.powerOnPasswd)
            rset = RadioSetting("settings.powerOnPasswd", "Power-on Password",
                                rs)
            function.append(rset)

            # Password
            def _char_to_str(chrx):
                """ Remove ff pads from char array """
                #  chrx is char array
                str1 = ""
                for sx in chrx:
                    if int(sx) > 31 and int(sx) < 127:
                        str1 += chr(int(sx))
                return str1

            def _pswd_vfy(setting, obj, atrb):
                """ Verify password is 1-6 chars, numbers 1-5 """
                str1 = str(setting.value).strip()  # initial
                # valid chars
                str2 = ''.join([c for c in str1 if c in '0123456789'])
                if str1 != str2:
                    # Two lines due to python 73 char limit
                    sx = "Bad characters in Password"
                    raise errors.RadioError(sx)
                str2 = str1.ljust(6, chr(00))  # pad to 6 with 00's
                setattr(obj, atrb, str2)
                return

            sx = _char_to_str(_password.digits).strip()
            rx = RadioSettingValueString(0, 6, sx)
            sx = "Password (numerals 0-9)"
            rset = RadioSetting("password.digits", sx, rx)
            rset.set_apply_callback(_pswd_vfy, _password, "digits")
            function.append(rset)

        # Menu 9 - Auto Power On
        rs = RadioSettingValueBoolean(_settings.autoPowerOn)
        rset = RadioSetting("settings.autoPowerOn", "Auto power on", rs)
        function.append(rset)

        # Menu 13 - Auto Power Off
        options = ["Off", "30 minutes", "60 minutes", "120 minutes"]
        rs = RadioSettingValueList(options, options[_settings.autoPowerOff])
        rset = RadioSetting("settings.autoPowerOff", "Auto power off", rs)
        function.append(rset)

        # Power On Reset Enable
        rs = RadioSettingValueBoolean(_settings.powerOnReset)
        rset = RadioSetting("settings.powerOnReset", "Power on reset", rs)
        function.append(rset)

        # FUNCTION SET
        # Menu 4 - Squelch Level A
        options = ["OFF"] + ["%s" % x for x in range(1, 10)]
        rs = RadioSettingValueList(options, options[_settings.squelchLevelA])
        rset = RadioSetting("settings.squelchLevelA", "Squelch level A", rs)
        function.append(rset)

        # Squelch Level B
        options = ["OFF"] + ["%s" % x for x in range(1, 10)]
        rs = RadioSettingValueList(options, options[_settings.squelchLevelB])
        rset = RadioSetting("settings.squelchLevelB", "Squelch level B", rs)
        function.append(rset)

        # Menu 7 - Scan Type
        options = ["Time operated (TO)", "Carrier operated (CO)",
                   "Search (SE)"]
        rs = RadioSettingValueList(options, options[_settings.scanType])
        rset = RadioSetting("settings.scanType", "Scan mode", rs)
        function.append(rset)

        # Menu 8 - Scan Recovery Time
        options = ["%s seconds" % x for x in range(5, 20, 5)]
        rs = RadioSettingValueList(options, options[_settings.scanRecoveryT])
        rset = RadioSetting("settings.scanRecoveryT", "Scan recovery time", rs)
        function.append(rset)

        # Main
        options = ["A", "B"]
        rs = RadioSettingValueList(options, options[_settings.main])
        rset = RadioSetting("settings.main", "Main", rs)
        function.append(rset)

        # Menu 10 - Dual Watch (RX Way Select)
        rs = RadioSettingValueBoolean(_settings.dualWatch)
        rset = RadioSetting("settings.dualWatch", "Dual watch", rs)
        function.append(rset)

        # Menu 12 - Time Out Timer
        options = ["OFF"] + ["%s minutes" % x for x in range(1, 31)]
        rs = RadioSettingValueList(options, options[_settings.timeOutTimer])
        rset = RadioSetting("settings.timeOutTimer", "Time out timer", rs)
        function.append(rset)

        # TBST Frequency
        options = ["1000 Hz", "1450 Hz", "1750 Hz", "2100 Hz"]
        rs = RadioSettingValueList(options, options[_settings.tbstFrequency])
        rset = RadioSetting("settings.tbstFrequency", "TBST frequency", rs)
        function.append(rset)

        # Save Channel Parameter
        rs = RadioSettingValueBoolean(_settings.saveChParameter)
        rset = RadioSetting("settings.saveChParameter",
                            "Save channel parameter", rs)
        function.append(rset)

        # MON Key Function
        options = ["Squelch off momentary", "Squelch off"]
        rs = RadioSettingValueList(options, options[_settings.monKeyFunction])
        rset = RadioSetting("settings.monKeyFunction", "MON key function", rs)
        function.append(rset)

        # Frequency Step
        options = ["2.5 kHz", "5 kHz", "6.25 kHz", "10 kHz", "12.5 kHz",
                   "20 kHz", "25 kHz", "30 kHz", "50 kHz"]
        rs = RadioSettingValueList(options, options[_settings.frequencyStep])
        rset = RadioSetting("settings.frequencyStep", "Frequency step", rs)
        function.append(rset)

        # Knob Mode
        options = ["Volume", "Channel"]
        rs = RadioSettingValueList(options, options[_settings.knobMode])
        rset = RadioSetting("settings.knobMode", "Knob mode", rs)
        function.append(rset)

        # TRF Enable
        rs = RadioSettingValueBoolean(_settings.trfEnable)
        rset = RadioSetting("settings.trfEnable", "TRF enable", rs)
        function.append(rset)

        if self.HAS_VOX:
            # VOX On/Off
            rs = RadioSettingValueBoolean(_settings.voxOnOff)
            rset = RadioSetting("settings.voxOnOff",
                                "VOX", rs)
            function.append(rset)

            # VOX Delay
            options = ["0.5 S", "1.0 S", "1.5 S", "2.0 S", "2.5 S",
                       "3.0 S", "3.5 S", "4.0 S", "4.5 S"]
            rs = RadioSettingValueList(options, options[_settings.voxDelay])
            rset = RadioSetting("settings.voxDelay", "VOX delay", rs)
            function.append(rset)

            # VOX Level
            options = ["%s" % x for x in range(1, 10)]
            rs = RadioSettingValueList(options, options[_settings.voxLevel])
            rset = RadioSetting("settings.voxLevel", "VOX Level", rs)
            function.append(rset)

        # Key Assignment
        pfkeys = RadioSettingGroup("pfkeys", "Key Assignment")
        group.append(pfkeys)

        options = ["A/B", "V/M", "SQL", "VOL", "POW", "CDT", "REV", "SCN",
                   "CAL", "TALK", "BND", "SFT", "MON", "DIR", "TRF", "RDW",
                   "NULL"]

        if self.HAS_VOX:
            options.insert(16, "VOX")

        # Key Mode 1
        # P1
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P1 - 1])
        rset = RadioSetting("pfkeys.keyMode1P1",
                            "Key mode 1 P1", rs)
        pfkeys.append(rset)

        # P2
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P2 - 1])
        rset = RadioSetting("pfkeys.keyMode1P2",
                            "Key mode 1 P2", rs)
        pfkeys.append(rset)

        # P3
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P3 - 1])
        rset = RadioSetting("pfkeys.keyMode1P3",
                            "Key mode 1 P3", rs)
        pfkeys.append(rset)

        # P4
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P4 - 1])
        rset = RadioSetting("pfkeys.keyMode1P4",
                            "Key mode 1 P4", rs)
        pfkeys.append(rset)

        # P5
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P5 - 1])
        rset = RadioSetting("pfkeys.keyMode1P5",
                            "Key mode 1 P5", rs)
        pfkeys.append(rset)

        # P6
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode1P6 - 1])
        rset = RadioSetting("pfkeys.keyMode1P6",
                            "Key mode 1 P6", rs)
        pfkeys.append(rset)

        # Key Mode 2
        # P1
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P1 - 1])
        rset = RadioSetting("pfkeys.keyMode2P1",
                            "Key mode 2 P1", rs)
        pfkeys.append(rset)

        # P2
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P2 - 1])
        rset = RadioSetting("pfkeys.keyMode2P2",
                            "Key mode 2 P2", rs)
        pfkeys.append(rset)

        # P3
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P3 - 1])
        rset = RadioSetting("pfkeys.keyMode2P3",
                            "Key mode 2 P3", rs)
        pfkeys.append(rset)

        # P4
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P4 - 1])
        rset = RadioSetting("pfkeys.keyMode2P4",
                            "Key mode 2 P4", rs)
        pfkeys.append(rset)

        # P5
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P5 - 1])
        rset = RadioSetting("pfkeys.keyMode2P5",
                            "Key mode 2 P5", rs)
        pfkeys.append(rset)

        # P6
        rs = RadioSettingValueList(options, options[_pfkeys.keyMode2P6 - 1])
        rset = RadioSetting("pfkeys.keyMode2P6",
                            "Key mode 2 P6", rs)
        pfkeys.append(rset)

        options = ["V/M", "SQL", "VOL", "POW", "CDT", "REV", "SCN", "CAL",
                   "TALK", "BND", "SFT", "MON", "DIR", "TRF", "RDW"]

        if self.HAS_VOX:
            options.insert(15, "VOX")

        # PA
        rs = RadioSettingValueList(options, options[_settings.keyPA - 2])
        rset = RadioSetting("settings.keyPA",
                            "Key PA", rs)
        pfkeys.append(rset)

        # PB
        rs = RadioSettingValueList(options, options[_settings.keyPB - 2])
        rset = RadioSetting("settings.keyPB",
                            "Key PB", rs)
        pfkeys.append(rset)

        # PC
        rs = RadioSettingValueList(options, options[_settings.keyPC - 2])
        rset = RadioSetting("settings.keyPC",
                            "Key PC", rs)
        pfkeys.append(rset)

        # PD
        rs = RadioSettingValueList(options, options[_settings.keyPD - 2])
        rset = RadioSetting("settings.keyPD",
                            "Key PD", rs)
        pfkeys.append(rset)

        # DTMF
        dtmf = RadioSettingGroup("dtmf", "DTMF")
        group.append(dtmf)

        # DTMF Transmitting Time
        options = ["50 milliseconds", "100 milliseconds", "200 milliseconds",
                   "300 milliseconds", "500 milliseconds"]
        rs = RadioSettingValueList(options, options[_settings.dtmfTxTime])
        rset = RadioSetting("settings.dtmfTxTime",
                            "DTMF transmitting time", rs)
        dtmf.append(rset)

        # DTMF Self ID

        # DTMF Interval Character
        IC_CHOICES = ["A", "B", "C", "D", "*", "#"]
        IC_VALUES = [0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F]

        def apply_ic_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) + " from list")
            val = str(setting.value)
            index = IC_CHOICES.index(val)
            val = IC_VALUES[index]
            obj.set_value(val)

        if _dtmf.intervalChar in IC_VALUES:
            idx = IC_VALUES.index(_dtmf.intervalChar)
        else:
            idx = IC_VALUES.index(0x0E)
        rs = RadioSetting("dtmf.intervalChar", "DTMF interval character",
                          RadioSettingValueList(IC_CHOICES,
                                                IC_CHOICES[idx]))
        rs.set_apply_callback(apply_ic_listvalue, _dtmf.intervalChar)
        dtmf.append(rs)

        # Group Code
        GC_CHOICES = ["Off", "A", "B", "C", "D", "*", "#"]
        GC_VALUES = [0xFF, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F]

        def apply_gc_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(setting.value) + " from list")
            val = str(setting.value)
            index = GC_CHOICES.index(val)
            val = GC_VALUES[index]
            obj.set_value(val)

        if _dtmf.groupCode in GC_VALUES:
            idx = GC_VALUES.index(_dtmf.groupCode)
        else:
            idx = GC_VALUES.index(0x0A)
        rs = RadioSetting("dtmf.groupCode", "DTMF interval character",
                          RadioSettingValueList(GC_CHOICES,
                                                GC_CHOICES[idx]))
        rs.set_apply_callback(apply_gc_listvalue, _dtmf.groupCode)
        dtmf.append(rs)

        # Decoding Response
        options = ["None", "Beep tone", "Beep tone & respond"]
        rs = RadioSettingValueList(options, options[_dtmf.decodingResponse])
        rset = RadioSetting("dtmf.decodingResponse", "Decoding response", rs)
        dtmf.append(rset)

        # First Digit Time
        options = ["%s" % x for x in range(0, 2510, 10)]
        rs = RadioSettingValueList(options, options[_dtmf.firstDigitTime])
        rset = RadioSetting("dtmf.firstDigitTime", "First Digit Time(ms)", rs)
        dtmf.append(rset)

        # First Digit Time
        options = ["%s" % x for x in range(10, 2510, 10)]
        rs = RadioSettingValueList(options, options[_dtmf.pretime - 1])
        rset = RadioSetting("dtmf.pretime", "Pretime(ms)", rs)
        dtmf.append(rset)

        # Auto Reset Time
        options = ["%s" % x for x in range(0, 25100, 100)]
        rs = RadioSettingValueList(options, options[_dtmf.autoResetTime])
        rset = RadioSetting("dtmf.autoResetTime", "Auto Reset time(ms)", rs)
        dtmf.append(rset)

        # Time-Lapse After Encode
        options = ["%s" % x for x in range(10, 2510, 10)]
        rs = RadioSettingValueList(options, options[_dtmf.timeLapse - 1])
        rset = RadioSetting("dtmf.timeLapse",
                            "Time-lapse after encode(ms)", rs)
        dtmf.append(rset)

        # PTT ID Pause Time
        options = ["Off", "-", "-", "-", "-"] + [
                   "%s" % x for x in range(5, 76)]
        rs = RadioSettingValueList(options, options[_dtmf.pauseTime])
        rset = RadioSetting("dtmf.pauseTime", "PTT ID pause time(s)", rs)
        dtmf.append(rset)

        # Side Tone
        rs = RadioSettingValueBoolean(_dtmf.sideTone)
        rset = RadioSetting("dtmf.sideTone", "Side tone", rs)
        dtmf.append(rset)

        # PTT ID Starting
        DTMF_CHARS = "0123456789ABCD*# "
        _codeobj = _dtmf.pttIdStart
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 16, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("dtmf.pttIdStart", "PTT ID starting", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 16):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.pttIdStart = code
        rs.set_apply_callback(apply_code, _dtmf)
        dtmf.append(rs)

        # PTT ID Ending
        _codeobj = _dtmf.pttIdEnd
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 16, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("dtmf.pttIdEnd", "PTT ID ending", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 16):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.pttIdEnd = code
        rs.set_apply_callback(apply_code, _dtmf)
        dtmf.append(rs)

        # Remotely Kill
        _codeobj = _dtmf.remoteKill
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 16, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("dtmf.remoteKill", "Remotely kill", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 16):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.remoteKill = code
        rs.set_apply_callback(apply_code, _dtmf)
        dtmf.append(rs)

        # Remotely Stun
        _codeobj = _dtmf.remoteStun
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 16, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("dtmf.remoteStun", "Remotely stun", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 16):
                try:
                    code.append(DTMF_CHARS.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.remoteStun = code
        rs.set_apply_callback(apply_code, _dtmf)
        dtmf.append(rs)

        # DTMF Encode
        # M1 - M16
        for i in range(0, 16):
            _codeobj = self._memobj.pttid[i].code
            _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            val = RadioSettingValueString(0, 16, _code, False)
            val.set_charset(DTMF_CHARS)
            rs = RadioSetting("pttid/%i.code" % i,
                              "DTMF encode M%i" % (i + 1), val)

            def apply_code(setting, obj):
                code = []
                for j in range(0, 16):
                    try:
                        code.append(DTMF_CHARS.index(str(setting.value)[j]))
                    except IndexError:
                        code.append(0xFF)
                obj.code = code
            rs.set_apply_callback(apply_code, self._memobj.pttid[i])
            dtmf.append(rs)

        return group

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "timeLapse":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "pretime":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "backlightBr":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "micKeyBrite":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "speakerVol":
                        setattr(obj, setting, int(element.value) + 1)
                    elif "keyMode" in setting:
                        setattr(obj, setting, int(element.value) + 1)
                    elif "keyP" in setting:
                        setattr(obj, setting, int(element.value) + 2)
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise


# Original non-VOX models
if has_future:
    @directory.register
    class AnyTone778UV(AnyTone778UVBase):
        VENDOR = "AnyTone"
        MODEL = "778UV"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'AT778UV': ['V100', 'V200']}

    @directory.register
    class RetevisRT95(AnyTone778UVBase):
        VENDOR = "Retevis"
        MODEL = "RT95"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'RT95': ['V100']}

    @directory.register
    class CRTMicronUV(AnyTone778UVBase):
        VENDOR = "CRT"
        MODEL = "Micron UV"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'MICRON': ['V100']}

    @directory.register
    class MidlandDBR2500(AnyTone778UVBase):
        VENDOR = "Midland"
        MODEL = "DBR2500"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'DBR2500': ['V100']}

    @directory.register
    class YedroYCM04vus(AnyTone778UVBase):
        VENDOR = "Yedro"
        MODEL = "YC-M04VUS"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'YCM04UV': ['V100']}


class AnyTone778UVvoxBase(AnyTone778UVBase):
    '''AnyTone 778UV VOX, Retevis RT95 VOX and others'''
    NAME_LENGTH = 6
    HAS_VOX = True


# New VOX models
if has_future:
    @directory.register
    class AnyTone778UVvox(AnyTone778UVvoxBase):
        VENDOR = "AnyTone"
        MODEL = "778UV VOX"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'778UV-P': ['V100']}

    @directory.register
    class RetevisRT95vox(AnyTone778UVvoxBase):
        VENDOR = "Retevis"
        MODEL = "RT95 VOX"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'RT95-P': ['V100']}

    @directory.register
    class CRTMicronUVvox(AnyTone778UVvoxBase):
        VENDOR = "CRT"
        MODEL = "Micron UV V2"
        # Allowed radio types is a dict keyed by model of a list of version
        # strings
        ALLOWED_RADIO_TYPES = {'MICRONP': ['V100']}



"""
Radio driver for the Boblov X3 Plus Motorcycle Helmet Radio
"""
# Copyright 2018 Robert C Jennings <rcj4747@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import struct
import time

from datetime import datetime

from chirp import (
    bitwise,
    chirp_common,
    directory,
    errors,
    memmap,
    util,
)
from chirp.settings import (
    RadioSetting,
    RadioSettingGroup,
    RadioSettings,
    RadioSettingValueBoolean,
    RadioSettingValueInteger,
    RadioSettingValueList,
)

LOG = logging.getLogger(__name__)


@directory.register
class BoblovX3Plus(chirp_common.CloneModeRadio,
                   chirp_common.ExperimentalRadio):
    """Boblov X3 Plus motorcycle/cycling helmet radio"""

    VENDOR = 'Boblov'
    MODEL = 'X3Plus'
    BAUD_RATE = 9600
    CHANNELS = 16

    MEM_FORMAT = """
    #seekto 0x0010;
    struct {
        lbcd rxfreq[4];
        lbcd txfreq[4];
        lbcd rxtone[2];
        lbcd txtone[2];
        u8 unknown1:1,
           compander:1,
           scramble:1,
           skip:1,
           highpower:1,
           narrow:1,
           unknown2:1,
           bcl:1;
        u8 unknown3[3];
    } memory[16];
    #seekto 0x03C0;
    struct {
        u8 unknown1:4,
           voiceprompt:2,
           batterysaver:1,
           beep:1;
        u8 squelchlevel;
        u8 unknown2;
        u8 timeouttimer;
        u8 voxlevel;
        u8 unknown3;
        u8 unknown4;
        u8 voxdelay;
    } settings;
    """

    # Radio command data
    CMD_ACK = '\x06'
    CMD_IDENTIFY = '\x02'
    CMD_PROGRAM_ENTER = '.VKOGRAM'
    CMD_PROGRAM_EXIT = '\x62'  # 'b'
    CMD_READ = 'R'
    CMD_WRITE = 'W'

    BLOCK_SIZE = 0x08

    VOICE_LIST = ['Off', 'Chinese', 'English']
    TIMEOUTTIMER_LIST = ['Off', '30 seconds', '60 seconds', '90 seconds',
                         '120 seconds', '150 seconds', '180 seconds',
                         '210 seconds', '240 seconds', '270 seconds',
                         '300 seconds']
    VOXLEVEL_LIST = ['Off', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    VOXDELAY_LIST = ['1 seconds', '2 seconds',
                     '3 seconds', '4 seconds', '5 seconds']
    X3P_POWER_LEVELS = [chirp_common.PowerLevel('Low', watts=0.5),
                        chirp_common.PowerLevel('High', watts=2.00)]

    _memsize = 0x03F0
    _ranges = [
        (0x0000, 0x03F0),
    ]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = _(
            "The X3Plus driver is currently experimental.\n"
            "There are no known issues but you should proceed with"
            " caution.\n"
            "Please save an unedited copy of your first successful\n"
            "download to a CHIRP Radio Images (*.img) file.\n")
        return rp

    @classmethod
    def match_model(cls, filedata, filename):
        """Given contents of a stored file (@filedata), return True if
          this radio driver handles the represented model"""

        if len(filedata) != cls._memsize:
            return False

        return b'P310' in filedata[0x03D0:0x03D8]

    def get_features(self):
        """Return a RadioFeatures object for this radio"""

        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.valid_modes = ['NFM', 'FM']  # 12.5 kHz, 25 kHz.
        rf.valid_power_levels = self.X3P_POWER_LEVELS
        rf.valid_skips = ['', 'S']
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_duplexes = ['', '-', '+', 'split', 'off']
        rf.can_odd_split = True
        rf.has_rx_dtcs = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_cross_modes = [
            'Tone->Tone',
            'DTCS->',
            '->DTCS',
            'Tone->DTCS',
            'DTCS->Tone',
            '->Tone',
            'DTCS->DTCS']
        rf.has_tuning_step = False
        rf.has_bank = False
        rf.has_name = False
        rf.memory_bounds = (1, self.CHANNELS)
        rf.valid_bands = [(400000000, 470000000)]
        return rf

    def process_mmap(self):
        """Process a newly-loaded or downloaded memory map"""
        self._memobj = bitwise.parse(self.MEM_FORMAT, self._mmap)

    def sync_in(self):
        "Initiate a radio-to-PC clone operation"

        LOG.debug('Cloning from radio')
        status = chirp_common.Status()
        status.msg = 'Cloning from radio'
        status.cur = 0
        status.max = self._memsize
        self.status_fn(status)

        self._enter_programming_mode()

        data = ''
        for addr in range(0, self._memsize, self.BLOCK_SIZE):
            status.cur = addr + self.BLOCK_SIZE
            self.status_fn(status)

            block = self._read_block(addr, self.BLOCK_SIZE)
            data += block

            LOG.debug('Address: %04x', addr)
            LOG.debug(util.hexprint(block))

        self._exit_programming_mode()

        self._mmap = memmap.MemoryMap(data)
        self.process_mmap()

    def sync_out(self):
        "Initiate a PC-to-radio clone operation"

        LOG.debug('Upload to radio')
        status = chirp_common.Status()
        status.msg = 'Uploading to radio'
        status.cur = 0
        status.max = self._memsize
        self.status_fn(status)

        self._enter_programming_mode()

        for start_addr, end_addr in self._ranges:
            for addr in range(start_addr, end_addr, self.BLOCK_SIZE):
                status.cur = addr + self.BLOCK_SIZE
                self.status_fn(status)

                self._write_block(addr, self.BLOCK_SIZE)

        self._exit_programming_mode()

    def get_raw_memory(self, number):
        """Return a raw string describing the memory at @number"""
        return repr(self._memobj.memory[number - 1])

    @staticmethod
    def _decode_tone(val):
        val = int(val)
        if val == 16665:
            return '', None, None
        elif val >= 12000:
            return 'DTCS', val - 12000, 'R'
        elif val >= 8000:
            return 'DTCS', val - 8000, 'N'

        return 'Tone', val / 10.0, None

    @staticmethod
    def _encode_tone(memval, mode, value, pol):
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception('Internal error: invalid mode `%s`' % mode)

    def get_memory(self, number):
        """Return a Memory object for the memory at location @number"""
        try:
            rmem = self._memobj.memory[number - 1]
        except KeyError:
            raise errors.InvalidMemoryLocation('Unknown channel %s' % number)

        if number < 1 or number > self.CHANNELS:
            raise errors.InvalidMemoryLocation(
                'Channel number must be 1 and %s' % self.CHANNELS)

        mem = chirp_common.Memory()
        mem.number = number
        mem.freq = int(rmem.rxfreq) * 10

        # A blank (0 MHz) or 0xFFFFFFFF frequency is considered empty
        if mem.freq == 0 or (
                rmem.rxfreq.get_raw(asbytes=False) == '\xFF\xFF\xFF\xFF'):
            LOG.debug('empty channel %d', number)
            mem.freq = 0
            mem.empty = True
            return mem

        if rmem.txfreq.get_raw(asbytes=False) == '\xFF\xFF\xFF\xFF':
            mem.duplex = 'off'
            mem.offset = 0
        elif int(rmem.rxfreq) == int(rmem.txfreq):
            mem.duplex = ''
            mem.offset = 0
        else:
            mem.duplex = '-' if int(rmem.rxfreq) > int(rmem.txfreq) else '+'
            mem.offset = abs(int(rmem.rxfreq) - int(rmem.txfreq)) * 10

        mem.mode = 'NFM' if rmem.narrow else 'FM'
        mem.skip = 'S' if rmem.skip else ''
        mem.power = self.X3P_POWER_LEVELS[rmem.highpower]

        txtone = self._decode_tone(rmem.txtone)
        rxtone = self._decode_tone(rmem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        mem.extra = RadioSettingGroup('Extra', 'extra')
        mem.extra.append(RadioSetting('bcl', 'Busy Channel Lockout',
                                      RadioSettingValueBoolean(
                                          current=(not rmem.bcl))))
        mem.extra.append(RadioSetting('scramble', 'Scramble',
                                      RadioSettingValueBoolean(
                                          current=(not rmem.scramble))))
        mem.extra.append(RadioSetting('compander', 'Compander',
                                      RadioSettingValueBoolean(
                                          current=(not rmem.compander))))

        return mem

    def set_memory(self, memory):
        """Set the memory object @memory"""
        rmem = self._memobj.memory[memory.number - 1]

        if memory.empty:
            rmem.set_raw('\xFF' * (rmem.size() // 8))
            return

        rmem.rxfreq = memory.freq / 10

        set_txtone = True
        if memory.duplex == 'off':
            for i in range(0, 4):
                rmem.txfreq[i].set_raw('\xFF')
                # If receive only then txtone value should be none
                self._encode_tone(rmem.txtone, mode='', value=None, pol=None)
                set_txtone = False
        elif memory.duplex == 'split':
            rmem.txfreq = memory.offset / 10
        elif memory.duplex == '+':
            rmem.txfreq = (memory.freq + memory.offset) / 10
        elif memory.duplex == '-':
            rmem.txfreq = (memory.freq - memory.offset) / 10
        else:
            rmem.txfreq = memory.freq / 10

        txtone, rxtone = chirp_common.split_tone_encode(memory)
        if set_txtone:
            self._encode_tone(rmem.txtone, *txtone)
        self._encode_tone(rmem.rxtone, *rxtone)

        rmem.narrow = 'N' in memory.mode
        rmem.skip = memory.skip == 'S'

        for setting in memory.extra:
            # NOTE: Only three settings right now, all are inverted
            setattr(rmem, setting.get_name(), not int(setting.value))

    def get_settings(self):
        """
        Return a RadioSettings list containing one or more RadioSettingGroup
        or RadioSetting objects.  These represent general settings that can
        be adjusted on the radio.
        """
        cur = self._memobj.settings
        basic = RadioSettingGroup('basic', 'Basic Settings')
        rs = RadioSetting('squelchlevel', 'Squelch level',
                          RadioSettingValueInteger(
                              minval=0, maxval=9,
                              current=cur.squelchlevel))
        basic.append(rs)
        rs = RadioSetting('timeouttimer', 'Timeout timer',
                          RadioSettingValueList(
                            options=self.TIMEOUTTIMER_LIST,
                            current=self.TIMEOUTTIMER_LIST[cur.timeouttimer]))
        basic.append(rs)
        rs = RadioSetting('voiceprompt', 'Voice prompt',
                          RadioSettingValueList(
                              options=self.VOICE_LIST,
                              current=self.VOICE_LIST[cur.voiceprompt]))
        basic.append(rs)
        rs = RadioSetting('voxlevel', 'Vox level',
                          RadioSettingValueList(
                              options=self.VOXLEVEL_LIST,
                              current=self.VOXLEVEL_LIST[cur.voxlevel]))
        basic.append(rs)
        rs = RadioSetting('voxdelay', 'VOX delay',
                          RadioSettingValueList(
                              options=self.VOXDELAY_LIST,
                              current=self.VOXDELAY_LIST[cur.voxdelay]))
        basic.append(rs)
        basic.append(RadioSetting('batterysaver', 'Battery saver',
                                  RadioSettingValueBoolean(
                                      current=cur.batterysaver)))
        basic.append(RadioSetting('beep', 'Beep',
                                  RadioSettingValueBoolean(
                                      current=cur.beep)))
        return RadioSettings(basic)

    def set_settings(self, settings):
        """
        Accepts the top-level RadioSettingGroup returned from
        get_settings() and adjusts the values in the radio accordingly.
        This function expects the entire RadioSettingGroup hierarchy
        returned from get_settings().
        """
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if '.' in element.get_name():
                        bits = element.get_name().split('.')
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug('Using apply callback')
                        element.run_apply_callback()
                    else:
                        LOG.debug('Setting %s = %s', setting, element.value)
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    def _write(self, data, timeout=3):
        """
        Write data to the serial port and consume the echoed response

        The radio echos the data it is sent before replying.  Send the
        data to the radio, consume the reply, and ensure that the reply
        is the same as the data sent.
        """

        serial = self.pipe
        expected = len(data)
        resp = b''
        start = datetime.now()

        # LOG.debug('WRITE(%02d): %s', expected, util.hexprint(data).rstrip())
        serial.write(data)
        while True:
            if not expected:
                break
            rbytes = serial.read(expected)
            resp += rbytes
            expected -= len(rbytes)
            if (datetime.now() - start).seconds > timeout:
                raise errors.RadioError('Timeout while reading from radio')
        if resp != data:
            raise errors.RadioError('Echoed response did not match sent data')

    def _read(self, length, timeout=3):
        """Read data from the serial port"""

        resp = b''
        serial = self.pipe
        remaining = length
        start = datetime.now()

        if not remaining:
            return resp

        while True:
            rbytes = serial.read(remaining)
            resp += rbytes
            remaining -= len(rbytes)
            if not remaining:
                break
            if (datetime.now() - start).seconds > timeout:
                raise errors.RadioError('Timeout while reading from radio')
            time.sleep(0.1)

        # LOG.debug('READ(%02d):  %s', length, util.hexprint(resp).rstrip())
        return resp

    def _read_block(self, block_addr, block_size):

        LOG.debug('Reading block %04x...', block_addr)
        cmd = struct.pack('>cHb', self.CMD_READ, block_addr,
                          block_size)
        resp_prefix = self.CMD_WRITE + cmd[1:]

        try:
            msg = ('Failed to write command to radio for block '
                   'read at %04x' % block_addr)
            self._write(cmd)

            msg = ('Failed to read response from radio for block '
                   'read at %04x' % block_addr)
            response = self._read(len(cmd) + block_size)

            if response[:len(cmd)] != resp_prefix:
                raise errors.RadioError('Error reading block %04x, '
                                        'Command not returned.' % (block_addr))

            msg = ('Failed to write ACK to radio after block read at '
                   '%04x' % block_addr)
            self._write(self.CMD_ACK)

            msg = ('Failed to read ACK from radio after block read at '
                   '%04x' % block_addr)
            ack = self._read(1)
        except Exception:
            LOG.debug(msg, exc_info=True)
            raise errors.RadioError(msg)

        if ack != self.CMD_ACK:
            raise errors.RadioError('No ACK reading block '
                                    '%04x.' % (block_addr))

        return response[len(cmd):]

    def _write_block(self, block_addr, block_size):

        cmd = struct.pack('>cHb', self.CMD_WRITE, block_addr, block_size)
        data = self.get_mmap()[block_addr:block_addr + 8]

        LOG.debug('Writing Data:\n%s%s',
                  util.hexprint(cmd), util.hexprint(data))

        try:
            self._write(cmd + data)
            if self._read(1) != self.CMD_ACK:
                raise Exception('No ACK')
        except Exception:
            msg = 'Failed to send block to radio at %04x' % block_addr
            LOG.debug(msg, exc_info=True)
            raise errors.RadioError(msg)

    def _enter_programming_mode(self):

        LOG.debug('Entering programming mode')
        try:
            msg = 'Error communicating with radio entering programming mode.'
            self._write(self.CMD_PROGRAM_ENTER)
            time.sleep(0.5)
            ack = self._read(1)

            if not ack:
                raise errors.RadioError('No response from radio')
            elif ack != self.CMD_ACK:
                raise errors.RadioError('Radio refused to enter '
                                        'programming mode')

            msg = 'Error communicating with radio during identification'
            self._write(self.CMD_IDENTIFY)
            ident = self._read(8)

            if not ident.startswith('SMP558'):
                LOG.debug(util.hexprint(ident))
                raise errors.RadioError('Radio returned unknown ID string')

            msg = ('Error communicating with radio while querying '
                   'model identifier')
            self._write(self.CMD_ACK)

            msg = 'Error communicating with radio on final handshake'
            ack = self._read(1)

            if ack != self.CMD_ACK:
                raise errors.RadioError('Radio refused to enter programming '
                                        'mode failed on final handshake.')
        except Exception:
            LOG.debug(msg, exc_info=True)
            raise errors.RadioError(msg)

    def _exit_programming_mode(self):
        try:
            self._write(self.CMD_PROGRAM_EXIT)
        except Exception:
            msg = 'Radio refused to exit programming mode'
            LOG.debug(msg, exc_info=True)
            raise errors.RadioError(msg)
        LOG.debug('Exited programming mode')



# Copyright 2021 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp import errors
from chirp import memmap
from chirp import util
from chirp.settings import RadioSettingGroup, RadioSetting, RadioSettings, \
    RadioSettingValueList, RadioSettingValueString, RadioSettingValueBoolean, \
    RadioSettingValueInteger

LOG = logging.getLogger(__name__)

#
#  Chirp Driver for Retevis RT98 models: RT98V (136-174 MHz)
#                                        RT98U (400-490 MHz)
#                                        RT98W (66-88 MHz)
#
#
#
# Global Parameters
#
TONES = [62.5] + list(chirp_common.TONES)
TMODES = ['', 'Tone', 'DTCS']
DUPLEXES = ['', '+', '-']

TXPOWER_LOW = 0x00
TXPOWER_MED = 0x01
TXPOWER_HIGH = 0x02

DUPLEX_NOSPLIT = 0x00
DUPLEX_POSSPLIT = 0x01
DUPLEX_NEGSPLIT = 0x02

CHANNEL_WIDTH_12d5kHz = 0x00
CHANNEL_WIDTH_20kHz = 0x01
CHANNEL_WIDTH_25kHz = 0x02

TUNING_STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 30.0, 50.0]

POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("Mid", watts=10),
                chirp_common.PowerLevel("High", watts=15)]

PMR_POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=0.5), ]

FREENET_POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1), ]

PMR_FREQS = [446006250, 446018750, 446031250, 446043750,
             446056250, 446068750, 446081250, 446093750,
             446106250, 446118750, 446131250, 446143750,
             446156250, 446168750, 446181250, 446193750]

FREENET_FREQS = [149025000, 149037500, 149050000,
                 149087500, 149100000, 149112500]

CROSS_MODES = ["Tone->Tone", "DTCS->", "->DTCS", "Tone->DTCS", "DTCS->Tone",
               "->Tone", "DTCS->DTCS"]

LIST_STEP = [str(x) for x in TUNING_STEPS]
LIST_TIMEOUT = ["Off"] + ["%s min" % x for x in range(1, 31)]
LIST_APO = ["Off", "30 min", "1 hr", "2 hrs"]
LIST_SQUELCH = ["Off"] + ["Level %s" % x for x in range(1, 10)]
LIST_DISPLAY_MODE = ["Channel", "Frequency", "Name"]
LIST_AOP = ["Manual", "Auto"]
LIST_STE_TYPE = ["Off", "Silent", "120 Degree", "180 Degree", "240 Degree"]
LIST_STE_FREQ = ["Off", "55.2 Hz", "259.2 Hz"]
LIST_VFOMR = ["MR", "VFO"]
LIST_SCAN = ["TO", "CO", "SE"]

LIST_PRIORITY_CH = ["Off", "Priority Channel 1", "Priority Channel 2",
                    "Priority Channel 1 + Priority Channel 2"]

LIST_REVERT_CH = ["Selected", "Selected + TalkBack", "Priority Channel 1",
                  "Priority Channel 2", "Last Called", "Last Used",
                  "Priority Channel 1 + TalkBack",
                  "Priority Channel 2 + TalkBack"]

LIST_TIME50 = ["0.1", "0.2", "0.3", "0.4", "0.5",
               "0.6", "0.7", "0.8", "0.9", "1.0",
               "1.1", "1.2", "1.3", "1.4", "1.5",
               "1.6", "1.7", "1.8", "1.9", "2.0",
               "2.1", "3.2", "2.3", "2.4", "2.5",
               "2.6", "2.7", "2.8", "2.9", "3.0",
               "3.1", "3.2", "3.3", "3.4", "3.5",
               "3.6", "3.7", "3.8", "3.9", "4.0",
               "4.1", "4.2", "4.3", "4.4", "4.5",
               "4.6", "4.7", "4.8", "4.9", "5.0"]
LIST_TIME46 = LIST_TIME50[4:]

LIST_RT98V_MODES = ["FreeNet", "COM", "COMII"]
LIST_RT98U_MODES = ["PMR", "COM", "COMII"]
LIST_RT98W_MODES = ["", "", "", "", "", "", "COM"]

LIST_RT98V_FREQS = ["Rx(149 - 149.2 MHz) Tx(149 - 149.2 MHz)",
                    "Rx(136 - 174 MHz) Tx(136 - 174 MHz)",
                    "Rx(147 - 174 MHz) Tx(147 - 174 MHz)"]

LIST_RT98U_FREQS = ["Rx(446 - 446.2 MHz) Tx(446 - 446.2 MHz)",
                    "Rx(400 - 470 MHz) Tx(400 - 470 MHz)",
                    "Rx(450 - 470 MHz) Tx(450 - 470 MHz)"]

LIST_RT98W_FREQS = ["",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Rx(66 - 88 MHz) Tx(66 - 88 MHz)"]

#  RT98  memory map
#  section: 1  Channel Bank
#         description of channel bank (199 channels , range 1-199)
#         Each 32 Byte (0x20 hex)  record:
#  bytes:bit  type                 description
#  ---------------------------------------------------------------------------
#  4          bbcd freq[4]         receive frequency in packed binary coded
#                                  decimal
#  4          bbcd offset[4]       transceive offset in packed binary coded
#                                  decimal (note: +/- direction set by
#                                  'duplex' field)
#  1          u8 unknown0
#  1          u8
#   :1        reverse:1            reverse flag, 0=off, 1=on (reverses
#                                  transmit and receive frequencies)
#   :1        txoff:1              transmitt off flag, 0=transmit, 1=do not
#                                  transmit
#   :2        power:2              transmit power setting, value range 0-2,
#                                  0=low, 1=middle, 2=high
#   :2        duplex:2             duplex settings, 0=simplex, 1=plus (+)
#                                  offset, 2=minus(-) offset (see offset field)
#   :2        channel_width:2      channel spacing, 0=12.5kHz, 1=20kHz, 2=25kHz
#  1          u8
#   :2        unknown1:2
#   :1        talkaround:1         talkaround flag, 0=off, 1=on
#                                  (bypasses repeater)
#   :1        squelch_mode:1       squelch mode flag, 0=carrier, 1=ctcss/dcs
#   :1        rxdcsextra:1         use with rxcode for index of rx DCS to use
#   :1        rxinv:1              inverse DCS rx polarity flag, 0=N, 1=I
#   :1        txdcsextra:1         use with txcode for index of tx DCS to use
#   :1        txinv:1              inverse DCS tx polarity flag, 0=N, 1=I
#  1          u8
#   :4        unknown2:4
#   :2        rxtmode:2            rx tone mode, value range 0-2, 0=none,
#                                  1=CTCSS, 2=DCS  (ctcss tone in field rxtone)
#   :2        txtmode:2            tx tone mode, value range 0-2, 0=none,
#                                  1=CTCSS, 3=DCS  (ctcss tone in field txtone)
#  1          u8
#   :2        unknown3:2
#   :6        txtone:6             tx ctcss tone, menu index
#  1          u8
#   :2        unknown4:2
#   :6        rxtone:6             rx ctcss tone, menu index
#  1          u8 txcode            ?, not used for ctcss
#  1          u8 rxcode            ?, not used for ctcss
#  1          u8
#   :6        unknown5:6
#   :1        busychannellockout:1 busy channel lockout flag, 0=off, 1=enabled
#   :1        unknown6:1
#  6          char name[6]         6 byte char string for channel name
#  9          u8 unknown7[9]
#
MEM_FORMAT = """
// #seekto 0x0000;
struct {
  bbcd freq[4];
  bbcd offset[4];
  u8 unknown0;
  u8 reverse:1,
     tx_off:1,
     txpower:2,
     duplex:2,
     channel_width:2;
  u8 unknown1:2,
     talkaround:1,
     squelch_mode:1,
     rxdcsextra:1,
     rxinv:1,
     txdcsextra:1,
     txinv:1;
  u8 unknown2:4,
     rxtmode:2,
     txtmode:2;
  u8 unknown3:2,
     txtone:6;
  u8 unknown4:2,
     rxtone:6;
  u8 txcode;
  u8 rxcode;
  u8 unknown5:6,
     busychannellockout:1,
     unknown6:1;
  char name[6];
  u8 unknown7[9];
} memory[199];
"""

#  RT98  memory map
#  section: 2 and 3  Channel Set/Skip Flags
#
#    Channel Set (starts 0x3240) : Channel Set  bit is value 0 if a memory
#                                  location in the channel bank is active.
#    Channel Skip (starts 0x3260): Channel Skip bit is value 0 if a memory
#                                  location in the channel bank is active.
#
#    Both flag maps are a total 24 bytes in length, aligned on 32 byte records.
#    bit = 0 channel not set/skip,  1 is channel set/no skip
#
#    to index a channel:
#        cbyte = channel / 8 ;
#        cbit  = channel % 8 ;
#        setflag  = csetflag[cbyte].c[cbit] ;
#        skipflag = cskipflag[cbyte].c[cbit] ;
#
#    channel range is 1-199, range is 32 bytes (last 7 unknown)
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3240;
struct {
   bit c[8];
} csetflag[32];

// #seekto 0x3260;
struct {
   bit c[8];
} cskipflag[32];

"""

#  RT98  memory map
#  section: 4  Startup Label
#
#  bytes:bit  type                 description
#  ---------------------------------------------------------------------------
#  6          char start_label[6]  label displayed at startup (usually
#                                  your call sign)
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3300;
struct {
    char startname[6];
} slabel;
"""

#  RT98  memory map
#  section: 5, 6 and 7  Radio Options
#        used to set a number of radio options
#
# description of function setup options, starting at 0x3310 (settings3)
#
#  bytes:bit  type                 description
#  ---------------------------------------------------------------------------
#  1          u8
#   :6        unknown:6
#   :2        bandlimit_3310:2     frequency ranges, range 0-2,
#                                  0=freenet(vhf) or pmr(uhf), 1=com, 2=comii
#                   rt98v - 00 FreeNet Rx(149 - 149.2 MHz) Tx(149 - 149.2 MHz)
#                           01 COM     Rx(136 - 174 MHz) Tx(136 - 174 MHz)
#                           02 COMII   Rx(147 - 174 MHz) Tx(147 - 174 MHz)
#                   rt98u - 00 PMR     Rx(446 - 446.2 MHz) Tx(446 - 446.2 MHz)
#                           01 COM     Rx(400 - 470 MHz) Tx(400 - 470 MHz)
#                           02 COMII   Rx(450 - 470 MHz) Tx(450 - 470 MHz)
#  1          u8 ch_number;        channel number, range 1-199
#  1          u8
#   :7        unknown_3312:7
#   :1        vfomr:1              vfo/mr mode, range 0-1, 0=mr, 1=vfo
#                                  only S/N:1907***********
#
# description of function setup options, starting at 0x3340 (settings)
#
#  bytes:bit  type                   description
#  ---------------------------------------------------------------------------
#  1          u8
#   :4        unknown_3340:4
#   :4        tuning_step:4          tuning step, menu index value from 0-8
#                                    2.5, 5, 6.25, 10, 12.5, 20, 25, 30, 50
#  1          u8
#   :7        unknown_3341:7
#   :1        beep:1                 beep mode, range 0-1, 0=off, 1=on
#  1           u8
#   :3        unknown_3342:3
#   :5        timeout_timer:5        timeout timer, range off (no timeout),
#                                    1-30 minutes
#  1          u8
#   :6        unknown_3343:6
#   :2        auto_power_off:2       auto power off, range 0-3, off, 30min,
#                                    1hr, 2hr
#  1          u8
#   :4        unknown_3344:4
#   :4        squelch:4              squelch level, range off, 1-9
#  1          u8
#   :3        unknown_3345:3
#   :5        volume:5               volume level, range 1-30 (no zero)
#  1          u8
#   :6        unknown_3346:6
#   :2        scan_revive:2          scan revive method, range 0-2, 0=to,
#                                    1=co, 2=se
#                                    only S/N:1907***********
#  1          u8 unknown_3347
#  1          u8   0x3348 [12]
#   :6        unknown_3348:6
#   :2        display_mode           display mode, range 0-2, 0=channel,
#                                    1=frequency, 2=name
#  1           u8
#   :7        unknown_3349:7
#   :1        auto_power_on:1        auto power on, range 0-1, 0=manual,
#                                    1=auto
#  1          u8
#   :3        unknown_334A:3
#   :5        mic_gain:5             mic gain, range 1-30 (no zero)
#  1          u8
#   :5        unknown_334C:5
#   :3        ste_type:3             ste type, range 0-4, 0=off, 1=silent,
#                                    2=120degree, 3=180degree, 4=240degree
#  1          u8
#   :7        unknown_334D:7
#   :1        ste_frequency:1        ste frequency, range 0-2, 0=off,
#                                    1=55.2Hz, 2=259.2Hz
#  1          u8
#   :2        unknown_0x334E:2
#   :1        forbid_setting:1       forbid setting(optional function),
#                                    range 0-1, 0=disabled, 1=enabled
#   :1        forbid_initialize:1    forbid initialize operate, range 0-1,
#                                    0=enabled, 1=disabled (inverted)
#   :1        save_chan_param:1      save channel parameters, range 0-1,
#                                    0=disabled, 1=enabled
#   :1        forbid_chan_menu:1     forbid channel menu, range 0-1,
#                                    0=disabled, 1=enabled
#   :1        sql_key_function:1     sql key function, range 0-1,
#                                    0=squelch off momentary, 1=squelch off
#   :1        unknown:1
#
# description of function setup options, starting at 0x3380 (settings2)
#
#  bytes:bit  type                   description
#  ---------------------------------------------------------------------------
#  1          u8
#   :7        unknown_3380:7
#   :1        scan_mode:1            scan mode, range 0-1, 0=off, 1=on
#  1          u8
#   :6        unknown_3381:6
#   :2        priority_ch:2          priority channel, range 0-3, 0=off,
#                                    1=priority channel 1,
#                                    2=priority channel 2,
#                                    3=priority channel 1 + priority channel 2
#  1          u8 priority_ch1        priority channel 1 number, range 1-199
#  1          u8 priority_ch2        priority channel 2 number, range 1-199
#  1          u8
#   :4        unknown_3384:4
#   :4        revert_ch:4            revert channel, range 0-3, 0=selected,
#                                    1=selected + talkback, 2=last called,
#                                    3=last used
#  1          u8 look_back_time_a    look back time a, range 0-45
#  1          u8 look_back_time_b    look back time b, range 0-45
#  1          u8 dropout_delay_time  dropout delay time, range 0-49
#  1          u8 dwell_time          dwell time, range 0-49
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3310;
struct {
    u8 bandlimit;
    u8 ch_number;
    u8 unknown_3312:7,
       vfomr:1;
} settings3;
"""

MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3340;
struct {
  u8 unknown_3340:4,
     tuning_step:4;
  u8 unknown_3341:7,
     beep:1;
  u8 unknown_3342:3,
     timeout_timer:5;
  u8 unknown_3343:6,
     auto_power_off:2;
  u8 unknown_3344:4,
     squelch:4;
  u8 unknown_3345:3,
     volume:5;
  u8 unknown_3346:6,
     scan_resume:2;
  u8 unknown_3347;
  u8 unknown_3348:6,
     display_mode:2;
  u8 unknown_3349:7,
     auto_power_on:1;
  u8 unknown_334A:3,
     mic_gain:5;
  u8 unknown_334B;
  u8 unknown_334C:5,
     ste_type:3;
  u8 unknown_334D:6,
     ste_frequency:2;
  u8 unknown_334E:1,
     forbid_setting:1,
     unknown1:1,
     forbid_initialize:1,
     save_chan_param:1,
     forbid_chan_menu:1,
     sql_key_function:1,
     unknown2:1;
} settings;
"""

MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3380;
struct {
  u8 unknown_3380:7,
     scan_mode:1;
  u8 unknown_3381:6,
     priority_ch:2;
  u8 priority_ch1;
  u8 priority_ch2;
  u8 unknown_3384:4,
     revert_ch:4;
  u8 look_back_time_a;
  u8 look_back_time_b;
  u8 dropout_delay_time;
  u8 dwell_time;
} settings2;
"""

#  RT98  memory map
#  section: 8  Embedded Messages
#
#  bytes:bit  type                 description
#  ---------------------------------------------------------------------------
#  6          char radio_type[5]   radio type, vhf=rt98v, uhf=rt98u
#  2          u8 unknown1[2]
#  4          char mcu_version[4]  mcu version, [x.xx]
#  2          u8 unknown2[2]
#  1          u8 mode              rt98u mode: 0=pmr, 1=com, 2=comii
#                                  rt98v mode: 0=freenet, 1=com, 2=comii
#  1          u8 unknown3
#  10         u8 unused1[10]
#  4          u8 unknown4[4]
#  3          u8 unused2[3]
#  16         u8 unknown5[16]
#  10         char date_mfg[16]    date manufactured, [yyyy-mm-dd]
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x3D00;
struct {
char radio_type[7];
char mcu_version[4];
u8 unknown2[2];
u8 mode;
u8 unknown3;
u8 unused1[10];
u8 unknown4[4];
u8 unused2[3];
u8 unknown5[16];
char date_mfg[10];
} embedded_msg;
"""


# Format for the version messages returned by the radio
VER_FORMAT = '''
u8 hdr;
char model[7];
u8 bandlimit;
char version[6];
u8 ack;
'''


# Radio supports upper case and symbols
CHARSET_ASCII_PLUS = chirp_common.CHARSET_UPPER_NUMERIC + '- '

# Band limits as defined by the band byte in ver_response, defined in Hz, for
# VHF and UHF, used for RX and TX.
RT98V_BAND_LIMITS = {0x00: [(149000000, 149200000)],
                     0x01: [(136000000, 174000000)],
                     0x02: [(147000000, 174000000)]}

RT98U_BAND_LIMITS = {0x00: [(446000000, 446200000)],
                     0x01: [(400000000, 470000000)],
                     0x02: [(450000000, 470000000)]}

RT98W_BAND_LIMITS = {0x00: [(66000000, 88000000)],
                     0x01: [(66000000, 88000000)],
                     0x02: [(66000000, 88000000)],
                     0x03: [(66000000, 88000000)],
                     0x04: [(66000000, 88000000)],
                     0x05: [(66000000, 88000000)],
                     0x06: [(66000000, 88000000)]}


# Get band limits from a band limit value
def get_band_limits_Hz(radio_type, limit_value):
    if str(radio_type).rstrip("\00") in ["RT98U", "AT-779U"]:
        if limit_value not in RT98U_BAND_LIMITS:
            limit_value = 0x01
            LOG.warning('Unknown band limit value 0x%02x, default to 0x01')
        bandlimitfrequencies = RT98U_BAND_LIMITS[limit_value]
    elif str(radio_type).rstrip("\00") in ["RT98V", "AT-779V"]:
        if limit_value not in RT98V_BAND_LIMITS:
            limit_value = 0x01
            LOG.warning('Unknown band limit value 0x%02x, default to 0x01')
        bandlimitfrequencies = RT98V_BAND_LIMITS[limit_value]
    elif str(radio_type).rstrip("\00") in ["RT98W", "AT-779W"]:
        if limit_value not in RT98W_BAND_LIMITS:
            limit_value = 0x06
            LOG.warning('Unknown band limit value 0x%02x, default to 0x06')
        bandlimitfrequencies = RT98W_BAND_LIMITS[limit_value]
    return bandlimitfrequencies


def _echo_write(radio, data):
    try:
        radio.pipe.write(data)
        radio.pipe.read(len(data))
    except Exception as e:
        LOG.error("Error writing to radio: %s" % e)
        raise errors.RadioError("Unable to write to radio")


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _read(radio, length):
    try:
        data = radio.pipe.read(length)
    except Exception as e:
        _finish(radio)
        LOG.error("Error reading from radio: %s" % e)
        raise errors.RadioError("Unable to read from radio")

    if len(data) != length:
        _finish(radio)
        LOG.error("Short read from radio (%i, expected %i)" %
                  (len(data), length))
        LOG.debug(util.hexprint(data))
        raise errors.RadioError("Short read from radio")
    return data


# strip trailing 0x00 to convert a string returned by bitwise.parse into a
# python string
def cstring_to_py_string(cstring):
    return "".join(c for c in cstring if c != '\x00')


# Check the radio version reported to see if it's one we support,
# returns bool version supported, and the band index
def check_ver(ver_response, allowed_types):
    ''' Check the returned radio version is one we approve of '''

    LOG.debug('ver_response = ')
    LOG.debug(util.hexprint(ver_response))

    resp = bitwise.parse(VER_FORMAT, ver_response)
    verok = False

    if resp.hdr == 0x49 and resp.ack == 0x06:
        model, version = [cstring_to_py_string(bitwise.get_string(s)).strip()
                          for s in (resp.model, resp.version)]
        LOG.debug('radio model: \'%s\' version: \'%s\'' %
                  (model, version))
        LOG.debug('allowed_types = %s' % allowed_types)

        if model in allowed_types:
            LOG.debug('model in allowed_types')

            if version in allowed_types[model]:
                LOG.debug('version in allowed_types[model]')
                verok = True
    else:
        _finish(radio)
        raise errors.RadioError('Failed to parse version response')

    return verok, str(resp.model), int(resp.bandlimit)


def _ident(radio):
    radio.pipe.timeout = 1
    _echo_write(radio, b"PROGRAM")
    response = radio.pipe.read(3)
    if response != b"QX\06":
        _finish(radio)
        LOG.debug("Response was :\n%s" % util.hexprint(response))
        raise errors.RadioError("Radio did not respond. Check connection.")
    _echo_write(radio, b"\x02")
    ver_response = radio.pipe.read(16)
    LOG.debug(util.hexprint(ver_response))

    verok, model, bandlimit = check_ver(ver_response,
                                        radio.ALLOWED_RADIO_TYPES)
    if not verok:
        _finish(radio)
        raise errors.RadioError(
            'Radio version not in allowed list for %s-%s: %s' %
            (radio.VENDOR, radio.MODEL, util.hexprint(ver_response)))

    return model, bandlimit


def _send(radio, cmd, addr, length, data=None):
    frame = struct.pack(">cHb", cmd, addr, length)
    if data:
        frame += data
        frame += bytes([_checksum(frame[1:])])
        frame += b"\x06"
    _echo_write(radio, frame)
    LOG.debug("Sent:\n%s" % util.hexprint(frame))
    if data:
        result = radio.pipe.read(1)
        if result != b"\x06":
            _finish(radio)
            LOG.debug("Ack was: %s" % repr(result))
            raise errors.RadioError("Radio did not accept block at %04x"
                                    % addr)
        return
    result = _read(radio, length + 6)
    LOG.debug("Got:\n%s" % util.hexprint(result))
    header = result[:4]
    data = result[4:-2]
    ack = result[-1:]
    if ack != b"\x06":
        _finish(radio)
        LOG.debug("Ack was: %s" % repr(ack))
        raise errors.RadioError("Radio NAK'd block at %04x" % addr)
    _cmd, _addr, _length = struct.unpack(">cHb", header)
    if _addr != addr or _length != _length:
        _finish(radio)
        LOG.debug("Expected/Received:")
        LOG.debug(" Length: %02x/%02x" % (length, _length))
        LOG.debug(" Addr: %04x/%04x" % (addr, _addr))
        raise errors.RadioError("Radio send unexpected block")
    cs = _checksum(result[1:-2])
    if cs != result[-2]:
        _finish(radio)
        LOG.debug("Calculated: %02x" % cs)
        LOG.debug("Actual:     %02x" % result[-2])
        raise errors.RadioError("Block at 0x%04x failed checksum" % addr)
    return data


def _finish(radio):
    endframe = b"\x45\x4E\x44"
    _echo_write(radio, endframe)
    result = radio.pipe.read(1)
    if result != b"\x06":
        LOG.error("Got:\n%s" % util.hexprint(result))
        raise errors.RadioError("Radio did not finish cleanly")


def do_download(radio):

    _ident(radio)

    _memobj = None
    data = b""

    for addr in range(0, radio._memsize, 0x10):
        block = _send(radio, b'R', addr, 0x10)
        data += block
        status = chirp_common.Status()
        status.cur = len(data)
        status.max = radio._memsize
        status.msg = "Downloading from radio"
        radio.status_fn(status)

    _finish(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    model, bandlimit = _ident(radio)
    _embedded = radio._memobj.embedded_msg

    if model != str(_embedded.radio_type):
        LOG.warning('radio and image model types differ')
        LOG.warning('model type (radio): %s' % str(model))
        LOG.warning('model type (image): %s' % str(_embedded.radio_type))

        _finish(radio)

        msg = ("The upload was stopped because the radio type "
               "of the image (%s) does not match that "
               "of the radio (%s).")
        raise errors.RadioError(msg % (str(_embedded.radio_type), str(model)))

    if bandlimit != int(_embedded.mode):
        if str(_embedded.radio_type).rstrip("\00") in ["RT98U", "AT-779U"]:
            image_band_limits = LIST_RT98U_FREQS[int(_embedded.mode)]
        if str(_embedded.radio_type).rstrip("\00") in ["RT98V", "AT-779V"]:
            image_band_limits = LIST_RT98V_FREQS[int(_embedded.mode)]
        if str(_embedded.radio_type).rstrip("\00") in ["RT98W", "AT-779W"]:
            image_band_limits = LIST_RT98W_FREQS[int(_embedded.mode)]
        if str(model).rstrip("\00") in ["RT98U", "AT-779U"]:
            radio_band_limits = LIST_RT98U_FREQS[int(bandlimit)]
        if str(model).rstrip("\00") in ["RT98V", "AT-779V"]:
            radio_band_limits = LIST_RT98V_FREQS[int(bandlimit)]
        if str(model).rstrip("\00") in ["RT98W", "AT-779W"]:
            radio_band_limits = LIST_RT98W_FREQS[int(bandlimit)]

        LOG.warning('radio and image band limits differ')
        LOG.warning('image band limits: %s' % image_band_limits)
        LOG.warning('radio band limits: %s' % radio_band_limits)

        _finish(radio)

        msg = ("The upload was stopped because the band limits "
               "of the image (%s) does not match that "
               "of the radio (%s).")
        raise errors.RadioError(msg % (image_band_limits, radio_band_limits))

    try:
        for start, end in radio._ranges:
            for addr in range(start, end, 0x10):
                block = radio._mmap[addr:addr+0x10]
                _send(radio, b'W', addr, len(block), block)
                status = chirp_common.Status()
                status.cur = addr
                status.max = end
                status.msg = "Uploading to Radio"
                radio.status_fn(status)
        _finish(radio)
    except errors.RadioError:
        raise
    except Exception as e:
        _finish(radio)
        raise errors.RadioError('Failed to upload to radio: %s' % e)


#
# The base class, extended for use with other models
#
class Rt98BaseRadio(chirp_common.CloneModeRadio,
                    chirp_common.ExperimentalRadio):
    """Retevis RT98 Base"""
    VENDOR = "Retevis"
    MODEL = "RT98 Base"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    _memsize = 0x3E00
    _ranges = [(0x0000, 0x3310),
               (0x3320, 0x3390)]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = ("The Retevis RT98 driver is an beta version."
                           "Proceed with Caution and backup your data")
        return rp

    def get_features(self):
        class FakeEmbedded(object):
            mode = 0
            radio_type = 'RT98U'

        if self._memobj:
            _embedded = self._memobj.embedded_msg
        else:
            # If we have no memory object, take defaults for unit
            # test, make_supported, etc
            _embedded = FakeEmbedded()
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.can_odd_split = True
        rf.has_name = True
        if _embedded.mode == 0:  # PMR or FreeNet
            rf.has_offset = False
        else:
            rf.has_offset = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_tuning_step = False
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.valid_skips = ["", "S"]
        rf.memory_bounds = (1, 199)
        rf.valid_name_length = 6
        if _embedded.mode == 0:  # PMR or FreeNet
            rf.valid_duplexes = ['', 'off']
        else:
            rf.valid_duplexes = DUPLEXES + ['split', 'off']
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "- "
        rf.valid_modes = ['FM', 'NFM']
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = CROSS_MODES
        if _embedded.mode == 0:  # PMR or FreeNet
            if str(_embedded.radio_type).rstrip("\00") == "RT98U":
                rf.valid_power_levels = PMR_POWER_LEVELS
            if str(_embedded.radio_type).rstrip("\00") == "RT98V":
                rf.valid_power_levels = FREENET_POWER_LEVELS
        else:
            rf.valid_power_levels = POWER_LEVELS
        rf.valid_tones = TONES
        rf.valid_dtcs_codes = chirp_common.ALL_DTCS_CODES

        try:
            rf.valid_bands = get_band_limits_Hz(
                str(_embedded.radio_type),
                int(_embedded.mode))
        except TypeError:
            # If we're asked without memory loaded, assume the most permissive
            rf.valid_bands = get_band_limits_Hz(str(_embedded.radio_type), 1)
        except Exception as e:
            LOG.error('Failed to get band limits for RT98: %s' % e)
            rf.valid_bands = get_band_limits_Hz(str(_embedded.radio_type), 1)

        rf.valid_tuning_steps = TUNING_STEPS
        return rf

    # Do a download of the radio from the serial port
    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    # Do an upload of the radio to the serial port
    def sync_out(self):
        do_upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    # Return a raw representation of the memory object, which
    # is very helpful for development
    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _get_dcs_index(self, _mem, which):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        return (int(extra) << 8) | int(base)

    def _set_dcs_index(self, _mem, which, index):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        base.set_value(index & 0xFF)
        extra.set_value(index >> 8)

    # Extract a high-level memory object from the low-level memory map
    # This is called to populate a memory in the UI
    def get_memory(self, number):
        _embedded = self._memobj.embedded_msg
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[number - 1]

        # get flag info
        cbyte = (number - 1) / 8
        cbit = 7 - ((number - 1) % 8)
        setflag = self._memobj.csetflag[cbyte].c[cbit]
        skipflag = self._memobj.cskipflag[cbyte].c[cbit]

        mem = chirp_common.Memory()

        mem.number = number  # Set the memory number

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if _mem.freq == 0:
            mem.empty = True
            return mem

        if setflag == 0:
            mem.empty = True
            return mem

        if _mem.get_raw()[0] == b"\xFF":
            mem.empty = True
            return mem

        # set the name
        mem.name = str(_mem.name).rstrip()  # Set the alpha tag

        # Convert your low-level frequency and offset to Hertz
        mem.freq = int(_mem.freq) * 10
        mem.offset = int(_mem.offset) * 10

        # Set the duplex flags
        if _mem.tx_off:  # handle tx off
            mem.duplex = 'off'
        elif _mem.duplex == DUPLEX_POSSPLIT:
            mem.duplex = '+'
        elif _mem.duplex == DUPLEX_NEGSPLIT:
            mem.duplex = '-'
        elif _mem.duplex == DUPLEX_NOSPLIT:
            mem.duplex = ''
        elif _mem.duplex == DUPLEX_ODDSPLIT:
            mem.duplex = 'split'
        else:
            LOG.error('%s: get_mem: unhandled duplex: %02x' %
                      (mem.name, _mem.duplex))

        # Set the channel width
        if _mem.channel_width == CHANNEL_WIDTH_12d5kHz:
            mem.mode = 'NFM'
        elif _embedded.mode == 0:  # PMR or FreeNet
            LOG.info('PMR and FreeNet channels must be Channel Width 12.5 kHz')
            mem.mode = 'NFM'
        elif _mem.channel_width == CHANNEL_WIDTH_25kHz:
            mem.mode = 'FM'
        elif _mem.channel_width == CHANNEL_WIDTH_20kHz:
            LOG.info(
                '%s: get_mem: promoting 20 kHz channel width to 25 kHz' %
                mem.name)
            mem.mode = 'FM'
        else:
            LOG.error('%s: get_mem: unhandled channel width: 0x%02x' %
                      (mem.name, _mem.channel_width))

        # set the power level
        if _embedded.mode == 0:  # PMR or FreeNet
            if str(_embedded.radio_type).rstrip("\00") == "RT98U":
                LOG.info('using PMR power levels')
                _levels = PMR_POWER_LEVELS
            if str(_embedded.radio_type).rstrip("\00") == "RT98V":
                LOG.info('using FreeNet power levels')
                _levels = FREENET_POWER_LEVELS
        else:  # COM or COMII
            LOG.info('using general power levels')
            _levels = POWER_LEVELS

        if _mem.txpower == TXPOWER_LOW:
            mem.power = _levels[0]
        elif _embedded.mode == 0:  # PMR or FreeNet
            LOG.info('FreeNet or PMR channel is not set to TX Power Low')
            LOG.info('Setting channel to TX Power Low')
            mem.power = _levels[0]
        elif _mem.txpower == TXPOWER_MED:
            mem.power = _levels[1]
        elif _mem.txpower == TXPOWER_HIGH:
            mem.power = _levels[2]
        else:
            LOG.error('%s: get_mem: unhandled power level: 0x%02x' %
                      (mem.name, _mem.txpower))

        # CTCSS Tones and DTCS Codes
        rxtone = txtone = None

        rxmode = TMODES[_mem.rxtmode]
        txmode = TMODES[_mem.txtmode]

        if rxmode == "Tone":
            rxtone = TONES[_mem.rxtone]
        elif rxmode == "DTCS":
            rxtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(
                                                 _mem, 'rx')]

        if txmode == "Tone":
            txtone = TONES[_mem.txtone]
        elif txmode == "DTCS":
            txtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(
                                                 _mem, 'tx')]

        rxpol = _mem.rxinv and "R" or "N"
        txpol = _mem.txinv and "R" or "N"

        chirp_common.split_tone_decode(mem,
                                       (txmode, txtone, txpol),
                                       (rxmode, rxtone, rxpol))

        # Check if this memory is in the scan enabled list
        mem.skip = "S" if skipflag == 0 else ""

        # Extra
        mem.extra = RadioSettingGroup("extra", "Extra")

        rs = RadioSettingValueBoolean(bool(_mem.busychannellockout))
        rset = RadioSetting("busychannellockout", "Busy channel lockout", rs)
        mem.extra.append(rset)

        rs = RadioSettingValueBoolean(bool(_mem.reverse))
        rset = RadioSetting("reverse", "Reverse", rs)
        mem.extra.append(rset)

        rs = RadioSettingValueBoolean(bool(_mem.talkaround))
        rset = RadioSetting("talkaround", "Talk around", rs)
        mem.extra.append(rset)

        rs = RadioSettingValueBoolean(bool(_mem.squelch_mode))
        rset = RadioSetting("squelch_mode", "Squelch mode", rs)
        mem.extra.append(rset)

        return mem

    # Store details about a high-level memory to the memory map
    # This is called when a user edits a memory in the UI
    def set_memory(self, mem):
        _embedded = self._memobj.embedded_msg
        # Get a low-level memory object mapped to the image

        _mem = self._memobj.memory[mem.number - 1]

        cbyte = (mem.number - 1) / 8
        cbit = 7 - ((mem.number - 1) % 8)

        if mem.empty:
            self._memobj.csetflag[cbyte].c[cbit] = 0
            self._memobj.cskipflag[cbyte].c[cbit] = 0
            _mem.set_raw("\xff" * 32)
            return

        _mem.set_raw("\x00" * 32)

        # FreeNet and PMR radio types
        if _embedded.mode == 0:  # PMR or FreeNet

            mem.mode = 'NFM'
            mem.offset = 0

            # FreeNet
            if str(_embedded.radio_type).rstrip("\00") == "RT98V":
                if mem.number >= 1 and mem.number <= 6:
                    FREENET_FREQ = FREENET_FREQS[mem.number - 1]
                    mem.freq = FREENET_FREQ
                else:
                    _mem.tx_off = 1
                    mem.duplex = 'off'

            # PMR
            if str(_embedded.radio_type).rstrip("\00") == "RT98U":
                if mem.number >= 1 and mem.number <= 16:
                    PMR_FREQ = PMR_FREQS[mem.number - 1]
                    mem.freq = PMR_FREQ
                else:
                    _mem.tx_off = 1
                    mem.duplex = 'off'

        # set the occupied bitfield
        self._memobj.csetflag[cbyte].c[cbit] = 1
        # set the scan add bitfield
        self._memobj.cskipflag[cbyte].c[cbit] = 0 if (mem.skip == "S") else 1

        _mem.freq = mem.freq / 10             # Convert to low-level frequency
        _mem.offset = mem.offset / 10         # Convert to low-level frequency

        # Store the alpha tag
        _mem.name = mem.name.ljust(6)[:6]  # Store the alpha tag

        # Set duplex bitfields
        _mem.tx_off = 0
        if mem.duplex == 'off':  # handle tx off
            _mem.tx_off = 1
        elif mem.duplex == '+':
            _mem.duplex = DUPLEX_POSSPLIT
        elif mem.duplex == '-':
            _mem.duplex = DUPLEX_NEGSPLIT
        elif mem.duplex == '':
            _mem.duplex = DUPLEX_NOSPLIT
        elif mem.duplex == 'split':
            diff = mem.offset - mem.freq
            _mem.duplex = DUPLEXES.index("-") \
                if diff < 0 else DUPLEXES.index("+")
            _mem.offset = abs(diff) / 10
        else:
            LOG.error('%s: set_mem: unhandled duplex: %s' %
                      (mem.name, mem.duplex))

        # Set the channel width - remember we promote 20 kHz channels to FM
        # on import, so don't handle them here
        if mem.mode == 'FM':
            _mem.channel_width = CHANNEL_WIDTH_25kHz
        elif mem.mode == 'NFM':
            _mem.channel_width = CHANNEL_WIDTH_12d5kHz
        else:
            LOG.error('%s: set_mem: unhandled mode: %s' % (
                mem.name, mem.mode))

        # CTCSS Tones and DTCS Codes
        ((txmode, txtone, txpol),
         (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(mem)

        _mem.txtmode = TMODES.index(txmode)

        _mem.rxtmode = TMODES.index(rxmode)

        if txmode == "Tone":
            _mem.txtone = TONES.index(txtone)
        elif txmode == "DTCS":
            self._set_dcs_index(_mem, 'tx',
                                chirp_common.ALL_DTCS_CODES.index(txtone))

        _mem.squelch_mode = False
        if rxmode == "Tone":
            _mem.rxtone = TONES.index(rxtone)
            _mem.squelch_mode = True
        elif rxmode == "DTCS":
            self._set_dcs_index(_mem, 'rx',
                                chirp_common.ALL_DTCS_CODES.index(rxtone))
            _mem.squelch_mode = True

        _mem.txinv = txpol == "R"
        _mem.rxinv = rxpol == "R"

        # set the power level
        if _embedded.mode == 0:  # PMR or FreeNet
            if str(_embedded.radio_type).rstrip("\00") == "RT98U":
                LOG.info('using PMR power levels')
                _levels = PMR_POWER_LEVELS
            if str(_embedded.radio_type).rstrip("\00") == "RT98V":
                LOG.info('using FreeNet power levels')
                _levels = FREENET_POWER_LEVELS
        else:  # COM or COMII
            LOG.info('using general power levels')
            _levels = POWER_LEVELS

        if mem.power is None:
            _mem.txpower = TXPOWER_HIGH
        elif mem.power == _levels[0]:
            _mem.txpower = TXPOWER_LOW
        elif _embedded.mode == 0:  # PMR or FreeNet
            LOG.info('FreeNet or PMR channel is not set to TX Power Low')
            LOG.info('Setting channel to TX Power Low')
            _mem.txpower = TXPOWER_LOW
        elif mem.power == _levels[1]:
            _mem.txpower = TXPOWER_MED
        elif mem.power == _levels[2]:
            _mem.txpower = TXPOWER_HIGH
        else:
            LOG.error('%s: set_mem: unhandled power level: %s' %
                      (mem.name, mem.power))

        # extra settings
        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def _get_settings(self):
        _embedded = self._memobj.embedded_msg
        _settings = self._memobj.settings
        _settings2 = self._memobj.settings2
        _settings3 = self._memobj.settings3
        _slabel = self._memobj.slabel

        function = RadioSettingGroup("function", "Function Setup")
        group = RadioSettings(function)

        # Function Setup
        # MODE SET
        rs = RadioSettingValueList(LIST_DISPLAY_MODE,
                                   LIST_DISPLAY_MODE[_settings.display_mode])
        rset = RadioSetting("display_mode", "Display Mode", rs)
        function.append(rset)

        if "AT-779" in str(_embedded.radio_type):
            rs = RadioSettingValueList(LIST_VFOMR,
                                       LIST_VFOMR[_settings3.vfomr])
            rset = RadioSetting("settings3.vfomr", "VFO/MR", rs)
            function.append(rset)

        rs = RadioSettingValueInteger(1, 199, _settings3.ch_number + 1)
        rset = RadioSetting("settings3.ch_number", "Channel Number", rs)
        function.append(rset)

        # DISPLAY SET
        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        val = RadioSettingValueString(0, 6, _filter(_slabel.startname))
        rs = RadioSetting("slabel.startname", "Startup Label", val)
        function.append(rs)

        # VOL SET
        rs = RadioSettingValueBoolean(bool(_settings.beep))
        rset = RadioSetting("beep", "Beep Prompt", rs)
        function.append(rset)

        rs = RadioSettingValueInteger(1, 30, _settings.volume)
        rset = RadioSetting("volume", "Volume Level", rs)
        function.append(rset)

        rs = RadioSettingValueInteger(1, 16, _settings.mic_gain)
        rset = RadioSetting("mic_gain", "Mic Gain", rs)
        function.append(rset)

        # ON/OFF SET
        rs = RadioSettingValueList(LIST_APO,
                                   LIST_APO[_settings.auto_power_off])
        rset = RadioSetting("auto_power_off", "Auto Power Off", rs)
        function.append(rset)

        rs = RadioSettingValueList(LIST_AOP, LIST_AOP[_settings.auto_power_on])
        rset = RadioSetting("auto_power_on", "Power On Method", rs)
        function.append(rset)

        # STE SET
        rs = RadioSettingValueList(LIST_STE_FREQ,
                                   LIST_STE_FREQ[_settings.ste_frequency])
        rset = RadioSetting("ste_frequency", "STE Frequency", rs)
        function.append(rset)

        rs = RadioSettingValueList(LIST_STE_TYPE,
                                   LIST_STE_TYPE[_settings.ste_type])
        rset = RadioSetting("ste_type", "STE Type", rs)
        function.append(rset)

        # FUNCTION SET
        rs = RadioSettingValueList(LIST_STEP, LIST_STEP[_settings.tuning_step])
        rset = RadioSetting("tuning_step", "Tuning Step", rs)
        function.append(rset)

        rs = RadioSettingValueList(LIST_SQUELCH,
                                   LIST_SQUELCH[_settings.squelch])
        rset = RadioSetting("squelch", "Squelch Level", rs)
        function.append(rset)

        if "AT-779" in str(_embedded.radio_type):
            rs = RadioSettingValueList(LIST_SCAN,
                                       LIST_SCAN[_settings.scan_resume])
            rset = RadioSetting("scan_resume", "Frequency Scan", rs)
            function.append(rset)

        rs = RadioSettingValueBoolean(bool(_settings.sql_key_function))
        rset = RadioSetting("sql_key_function", "SQL Key Function", rs)
        function.append(rset)

        rs = RadioSettingValueList(LIST_TIMEOUT,
                                   LIST_TIMEOUT[_settings.timeout_timer])
        rset = RadioSetting("timeout_timer", "Timeout Timer", rs)
        function.append(rset)

        # uncategorized
        rs = RadioSettingValueBoolean(bool(_settings.save_chan_param))
        rset = RadioSetting("save_chan_param", "Save Channel Parameters", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(bool(_settings.forbid_chan_menu))
        rset = RadioSetting("forbid_chan_menu", "Forbid Channel Menu", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(bool(not _settings.forbid_initialize))
        rset = RadioSetting("forbid_initialize", "Forbid Initialize", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(bool(_settings.forbid_setting))
        rset = RadioSetting("forbid_setting", "Forbid Setting", rs)
        function.append(rset)

        # Information Of Scanning Channel
        scanning = RadioSettingGroup("scanning", "Scanning Setup")
        group.append(scanning)

        rs = RadioSettingValueBoolean(bool(_settings2.scan_mode))
        rset = RadioSetting("settings2.scan_mode", "Scan Mode", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_PRIORITY_CH,
                                   LIST_PRIORITY_CH[_settings2.priority_ch])
        rset = RadioSetting("settings2.priority_ch", "Priority Channel", rs)
        scanning.append(rset)

        rs = RadioSettingValueInteger(1, 199, _settings2.priority_ch1 + 1)
        rset = RadioSetting("settings2.priority_ch1", "Priority Channel 1", rs)
        scanning.append(rset)

        rs = RadioSettingValueInteger(1, 199, _settings2.priority_ch2 + 1)
        rset = RadioSetting("settings2.priority_ch2", "Priority Channel 2", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_REVERT_CH,
                                   LIST_REVERT_CH[_settings2.revert_ch])
        rset = RadioSetting("settings2.revert_ch", "Revert Channel", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_TIME46,
                                   LIST_TIME46[_settings2.look_back_time_a])
        rset = RadioSetting("settings2.look_back_time_a",
                            "Look Back Time A", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_TIME46,
                                   LIST_TIME46[_settings2.look_back_time_b])
        rset = RadioSetting("settings2.look_back_time_b",
                            "Look Back Time B", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_TIME50,
                                   LIST_TIME50[_settings2.dropout_delay_time])
        rset = RadioSetting("settings2.dropout_delay_time",
                            "Dropout Delay Time", rs)
        scanning.append(rset)

        rs = RadioSettingValueList(LIST_TIME50,
                                   LIST_TIME50[_settings2.dwell_time])
        rset = RadioSetting("settings2.dwell_time", "Dwell Time", rs)
        scanning.append(rset)

        # Embedded Message
        embedded = RadioSettingGroup("embedded", "Embedded Message")
        group.append(embedded)

        rs = RadioSettingValueString(0, 7, _filter(_embedded.radio_type))
        rs.set_mutable(False)
        rset = RadioSetting("embedded_msg.radio_type", "Radio Type", rs)
        embedded.append(rset)

        if str(_embedded.radio_type).rstrip("\00") in ["RT98V", "AT-779V"]:
            options = LIST_RT98V_MODES
        elif str(_embedded.radio_type).rstrip("\00") in ["RT98W", "AT-779W"]:
            options = LIST_RT98W_MODES
        else:
            options = LIST_RT98U_MODES
        rs = RadioSettingValueList(options, options[_embedded.mode])
        rs.set_mutable(False)
        rset = RadioSetting("embedded_msg.mode", "Mode", rs)
        embedded.append(rset)

        # frequency
        if str(_embedded.radio_type).rstrip("\00") in ["RT98V", "AT-779V"]:
            options = LIST_RT98V_FREQS
        elif str(_embedded.radio_type).rstrip("\00") in ["RT98W", "AT-779W"]:
            options = LIST_RT98W_FREQS
        else:
            options = LIST_RT98U_FREQS
        rs = RadioSettingValueList(options, options[_settings3.bandlimit])
        rs.set_mutable(False)
        rset = RadioSetting("settings3.bandlimit", "Frequency", rs)
        embedded.append(rset)

        rs = RadioSettingValueString(0, 10, _filter(_embedded.date_mfg))
        rs.set_mutable(False)
        rset = RadioSetting("embedded_msg.date_mfg", "Production Date", rs)
        embedded.append(rset)

        rs = RadioSettingValueString(0, 4, _filter(_embedded.mcu_version))
        rs.set_mutable(False)
        rset = RadioSetting("embedded_msg.mcu_version", "MCU Version", rs)
        embedded.append(rset)

        return group

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("failed to parse settings")
            traceback.print_exc()
            return None

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("using apply callback")
                        element.run_apply_callback()
                    elif setting == "ch_number":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "forbid_initialize":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "priority_ch1":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "priority_ch2":
                        setattr(obj, setting, int(element.value) - 1)
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class Rt98Radio(Rt98BaseRadio):
    """Retevis RT98"""
    VENDOR = "Retevis"
    MODEL = "RT98"
    # Allowed radio types is a dict keyed by model of a list of version
    # strings
    ALLOWED_RADIO_TYPES = {'RT98V': ['V100'],
                           'RT98U': ['V100'],
                           'RT98W': ['V100'],
                           'AT-779V': ['V100'],
                           'AT-779U': ['V100'],
                           'AT-779W': ['V100'],
                           }
